# flake8: noqa
"""QM/MM interface with QM=FHI-aims, MM=gromacs

QM could be something else, but you need to read in qm-atom charges
from the qm program (in method 'get_qm_charges')


One can have many QM regions, each with a different calculator.
There can be only one MM calculator, which is calculating the whole
system.


Non-bonded interactions:
------------------------
Generally:

Within the same QM-QM:
  by qm calculator
MM-MM:
  by MM calculator
QM-MM:
  by MM using MM vdw parameters and QM charges.
Different QM different QM:
  by MM using QM and MM charges and MM-vdw parameters

The Hirschfeld charges (or other atomic charges)
on QM atoms are calculated by QM in a H terminated cluster in vacuum.
The charge of QM atom next to MM atom (edge-QM-atom)
and its H neighbors are set as in the classical force field.
The extra(missing) charge results from:

1) linkH atoms
2) The edge-QM atoms, and their qm-H neighbors,
   have their original MM charges.
3) and the fact that the charge of the QM fraction
   is not usually an integer when using the original MM charges.
   It is added equally to all QM atoms
   (not being linkH and not being edge-QM-atom or its H neighbor)
   so that the total charge of the MM-fragment involving QM atoms
   will be the same as in the original MM-description.

Vdw interactions are calculated by MM-gromacs for MM and MM-QM inteactions.
The QM-QM vdw interaction s could be done by the FHI-aims if desired
(by modifying the imput for QM-FHI-aims input accordingly.

Bonded interactions::

  E=
  E_qm(QM-H)         ; qm energy of H terminated QM cluster(s)
  + E_mm(ALL ATOMS)  ; mm energy of all atoms,
                     ; except for terms in which all MM-interacting atoms are
                     ; in the same QM region

Forces do not act on link atoms but they are positioned by scaling.
Forces on link atoms are given to their QM and MM neighbors by chain rule.
(see J. Chem. Theory Comput. 2011, 7, 761-777).
The optimal edge-qm-atom-linkH bond length is calculated
by QM in 'get_eq_qm_atom_link_h_distances'
or they are read from a file.


Questions & Comments markus.kaukonen@iki.fi

I'm especially interested in cases when we need two or more
QM regions. For instance two redox centers in a protein,
cathode and anode of a fuel cell ... you name it!


Some things to improve:

1) Water topology issue (at the moment water cannot be in QM),
   Its topology should be put into the main
   topology file, not in a separate file.

2) point charges and periodicity (if desired) to the QM calculation
   (now in vacuum)

3) Eichinger type of link atom treatment with fitted force constants for
   linkH-QMedge (bond strecth)
   linkH-QMedge-QMnextTOedge (angle terms)

4) file io using unformatted formats (.trr) instead of g96
   This is not easily possible without loading extra stuff from
   ftp://ftp.gromacs.org/pub/contrib/xd...e-1.1.1.tar.gz.

5) Utilize gromacs-python wrapper: (just found this today 31.12.2012...)
   http://orbeckst.github.com/GromacsWrapper/index.html#
"""

import sys
import numpy as np


def get_neighbor_list(system):
    """
    Makes a neighbor list of a system (ase Atoms).
    See
    https:\
    //wiki.fysik.dtu.dk/ase/ase/calculators/calculators.html#module-calculators
    """

    from ase.neighborlist import NeighborList
    from ase.data import covalent_radii
    import os
    import pickle

    NEIGHBOR_FILE = 'neighbor_list_for_ase_qmmm.txt'

    if os.path.exists(NEIGHBOR_FILE):
        print('Reading qm/mm neighbor list from file:')
        print('neighbor_list_for_ase_qmmm.txt')

        myfile = open(NEIGHBOR_FILE, 'rb')
        neighbor_list = pickle.load(myfile)
    else:
        cut = [covalent_radii[atom.number] for atom in system]
        skin = [0.2 for atom in system]
        
        neighbor_list = NeighborList(cut, skin, \
                                         self_interaction=False, bothways=True)
        neighbor_list.update(system)
        file = open(NEIGHBOR_FILE, 'wb')
        pickle.dump(neighbor_list, file)
        file.close()

    return neighbor_list

def get_qm_atoms(indexfilename='index.ndx'):
    """
    Read the indexes of all QM atoms (there may be many QM regions)
    """

    infile = open(indexfilename,'r')
    lines = infile.readlines()
    infile.close()
    qms = []

    for iline, line in enumerate(lines):
        if (('[ QM' in line) or ('[ qm' in line) or ('[ Qm' in line)) \
                or (('[QM' in line) or ('[qm' in line) or ('[Qm' in line)):
            qm = []
            for checkline in lines[iline+1:]:
                if ('[') in checkline:
                    break
                else:
                    qm = qm + [int(float(s)-1.0) for s in \
                                   checkline.split() if s.isdigit()]
            qm = list(set(qm))
            qms.append(qm)
    return qms

class LinkAtom:
    """
    Class for information about a single link-atom
    (it terminates a QM cluster)

    qm_region_index and link_atom_index refer to the following indexing system:
    [[QM0 link atoms indexes from 0],[QM1 link atoms indexes from 0],...]
    So above the second link atom in second qm region would have
    qm_region_index=1,  link_atom_index=1

    link_atom_index_in_qm tells which index in qm system the link atom has
    for instance
    qm_region_index=1, link_atom_index_in_qm=20
    means that link atom is 21'st atom in the second qm system

    """
    def __init__(self, atom, qm_region_index, link_atom_index):
        """ set initial values to a link atom object """
        self.atom = atom
        self.qm_region_index = qm_region_index
        self.link_atom_index = link_atom_index
        self.link_atom_index_in_qm = None
        self.qm_neighbor = None
        self.mm_neighbor = None
        self.qm2_neighbors = []
        self.qm3_neighbors = []
        self.mm2_neighbors = []
        self.set_qm2_neighbors = set([])
        self.set_qm3_neighbors = set([])
        self.set_mm2_neighbors = set([])
        self.force_constant = 0.0
        self.equilibrium_distance_xh = 0.0
        self.equilibrium_distance_xy = 0.0


    def set_link_atom(self, atom):
        """ set an ase-atom to be the link atom """
        self.atom = atom

    def set_link_atom_qm_region_index(self, qm_region_index):
        """ set to which qm region the link atom belongs to """
        self.qm_region_index = qm_region_index

    def set_link_atom_index_in_qm(self, link_atom_index_in_qm):
        """ set what is my link atom index in this qm region """
        self.link_atom_index_in_qm = link_atom_index_in_qm

    def set_link_atom_qm_neighbor(self, qm_neighbor):
        """ set what index does my qm neighbor have"""
        self.qm_neighbor = qm_neighbor

    def set_link_atom_mm_neighbor(self, mm_neighbor):
        """ set what index does my mm neighbor have"""
        self.mm_neighbor = mm_neighbor

    def set_link_atom_qm2_neighbors(self, qm2_neighbors):
        """ set what index does my second qm neighbor have"""
        self.qm2_neighbors = qm2_neighbors

    def set_link_atom_qm3_neighbors(self, qm3_neighbors):
        """ set what index does my third qm neighbor have"""
        self.qm3_neighbors = qm3_neighbors

    def set_link_atom_mm2_neighbors(self, mm2_neighbors):
        """ set what index does my second mm neighbor have"""
        self.mm2_neighbors = mm2_neighbors

    def set_force_constant(self, force_constant):
        """ set the force constant of bond edge-qm -- linkH (not used)"""
        self.force_constant = force_constant

    def set_equilibrium_distance_xh(self, equilibrium_distance_xh):
        """ set the equilibrium edge-qm -- linkH distance """
        self.equilibrium_distance_xh = equilibrium_distance_xh

    def set_equilibrium_distance_xy(self, equilibrium_distance_xy):
        """set the equilibrium edge-qm --
           edge-mm distance (by MM-force field)"""
        self.equilibrium_distance_xy = equilibrium_distance_xy

    def get_link_atom(self):
        """ get an ase-atom to be the link atom """
        return self.atom

    def get_link_atom_qm_region_index(self):
        """ get to which qm region the link atom belongs to """
        return self.qm_region_index

    def get_link_atom_index_in_qm(self):
        """ get what is my link atom index in this qm region """
        return self.link_atom_index_in_qm

    def get_link_atom_qm_neighbor(self):
        """ get what index does my qm neighbor have"""
        return self.qm_neighbor

    def get_link_atom_mm_neighbor(self):
        """ get what index does my mm neighbor have"""
        return self.mm_neighbor

    def get_link_atom_qm2_neighbors(self):
        """ get what index does my second qm neighbor have"""
        return self.qm2_neighbors

    def get_link_atom_qm3_neighbors(self):
        """ get what index does my third qm neighbor have"""
        return self.qm3_neighbors

    def get_link_atom_mm2_neighbors(self):
        """ get what index does my second mm neighbor have"""
        return self.mm2_neighbors

    def get_force_constant(self):
        """ get the force constant of bond edge-qm -- linkH (not used)"""
        return self.force_constant

    def get_equilibrium_distance_xh(self):
        """ get the equilibrium edge-qm -- linkH distance """
        return self.equilibrium_distance_xh

    def get_equilibrium_distance_xy(self):
        """get the equilibrium edge-qm --
           edge-mm distance (by MM-force field)"""
        return self.equilibrium_distance_xy


class AseQmmmManyqm:
    """ This is a qm/mm interface with qm=FHI-aims, mm=gromacs.
    
    We can have many QM regions, each with a different calculator.
    There can be only one MM calculator, which is calculating the whole
    system.

    Numeration of atoms starts from 0. (in qms, mms)

    In qm calculations link atom(s) come(s) last.

    For any qm region, the optimal bond lengths for all edge_atom-link_atom
    pairs are optimized by QM simultaneously at the beginning of
    the run when the flag link_info='byQM' is used (by method . The positions of other a

    """

    def __init__(self,  nqm_regions,  \
                     qm_calculators, mm_calculator, \
                     link_info='byQM'):
        """ Set initial values to each qm and mm calculator.
        Additionally set information for the qm/mm interface.

        The information about qm and mm indexes is read from
        a file 'index.ndx'
        Which can be generated with a gromacs tool 'make_ndx'
        http://www.gromacs.org/Documentation/Gromacs_Utilities/make_ndx

        Parameters
        ==========
        nqm_regions: int
            how many qm regions

        qm_calculators: list members of a Class defining a Calculator
            ase-qm calculator for each qm region

        mm_calculator: a member of a Class defining a Calculator
             ase-mm calculator for mm (the whole system)

        link_info: str
            can be either
            'byQM': the edge_qm_atom-link_h_atom distances are calculated by QM
            'byFile':the edge_qm_atom-link_h_atom distances are read from a file

        """

        from ase.io import write
        import os, glob

        # clean
        files = glob.glob('test-*')
        for file in files:
            try:
                os.remove(file)
            except OSError:
                pass

        self.atoms = None
        self.positions = None
        self.neighbor_list = None
        self.link_atoms = []

        self.energy = None
        self.e_delta_stretch = None

        self.nqm_regions = nqm_regions
        self.qm_calculators = qm_calculators
        self.mm_calculator = mm_calculator

        self.qmatom_types = []
        self.mmatom_types = []

        #det unique name for each qm region
        # (the output file of each qm calculation)
        #for i in range(len(self.qm_calculators)):
        #    self.qm_calculators[i].set(output_template = 'aims'+str(i))

        self.link_systems = None

        self.equilibrium_distances_xy = []
        self.equilibrium_distances_xh = []
        self.force_constants = []

        # get the sets of qm atoms
        self.qms = get_qm_atoms()
        self.set_qms = set(sum(self.qms, []))
        print('qmsystem(s), indexing from 0:')
        print('')
        for index_out in self.qms:
            index_str = ''
            for index in index_out:
                index_str += str(index) + ' '
            print('%s' % index_str)
            print('')

        if len(self.qms) != nqm_regions:
            print('Number of set of QM atoms does not match with nqm_regions')
            print('self.qms %s' % str(self.qms))
            print('nqm_regions %s' % str(nqm_regions))
            sys.exit()
        if len(self.qms) != len(qm_calculators):
            print('Number of set of QM atoms does not match with')
            print('the number of QM calculators')
            sys.exit()


        #read the actual structure to define link atoms and their neighbors
        system_tmp =  mm_calculator.atoms
        self.positions = system_tmp.get_positions()

        #get neighbor lists
        self.neighbor_list = get_neighbor_list(system_tmp)

        #get the mm-atoms next to link atoms for all qm regions
        (self.mms_edge, self.qms_edge, self.set_mms_edge, self.set_qms_edge) = \
            self.get_edge_qm_and_mm_atoms(self.qms, system_tmp)

        #get the mm atoms being second neighbors to any qm atom
        (self.second_mms, self.set_second_mms) = \
            self.get_next_neighbors(self.mms_edge, self.set_qms)

        #get the qm atoms being second neighbors to link atom
        (self.second_qms, self.set_second_qms) = \
            self.get_next_neighbors(self.qms_edge, \
                                        self.set_mms_edge)

        #get the qm atoms being  neighbors to link atom (edge-qm atoms)
        # and their neighbors which have only single neighbor
        # (for example edge-QM(C)-H or edge-QM(C)=O; for charge exclusion)
        self.constant_charge_qms = \
            self.get_constant_charge_qms\
            (self.set_qms_edge, self.set_second_qms)

        #get the qm atoms being third neighbors to link atom
        (self.third_qms, self.set_third_qms) = \
            self.get_next_neighbors\
            (self.second_qms, self.set_qms_edge)
        print('self.qms %s' % self.qms)
        print('QM edge, MM edge %s'
              % str(self.qms_edge)+' '+ str(self.mms_edge))
        print('MM second N of Link %s' % str(self.second_mms))
        print('QM second N of Link %s' % str(self.second_qms))
        print('QM third N of Link %s' % str(self.third_qms))

        #check that MM topology exists:
        if not(os.path.exists(mm_calculator.topology_filename)):
            print('NO TOPOLOGY FILE:', mm_calculator.topology_filename)
            print('use: CALC_MM.generate_topology_and_g96file()')
            sys.exit()

        #check that MM run file (.tpr) exists:
        if not(os.path.exists(mm_calculator.label+'.tpr')):
            print('NO MM run file FILE:', mm_calculator.label + '.tpr')
            print('use: CALC_MM.write_input(atoms)')
            print('use: CALC_MM.generate_gromacs_run_file()')
            sys.exit()
        #check that force field files exist
        if 'GMXDATA' in os.environ:
            gromacs_home = os.environ['GMXDATA'].split(':')[0]
        else:
            gromacs_home = '/usr/local/gromacs/share/gromacs/'
        ff_filename = gromacs_home+ '/top/' \
            + mm_calculator.force_field + '.ff/ffbonded.itp'
        if not(os.path.exists(mm_calculator.topology_filename)):
            print('NO force field file:', ff_filename)
            print('use: GMXDATA environmental variable')
            sys.exit()

        if link_info == 'byFILE':
            self.read_eq_distances_from_file()
        else:
            
            #get QM-MM bond lengths
            self.get_eq_distances_xy(\
                topfilename=mm_calculator.topology_filename,\
                    force_field= mm_calculator.force_field)

            #get QM-linkH distances by QM for all link atoms
            self.get_eq_qm_atom_link_h_distances(system_tmp)


        # write current link-info data to file (it can be later used,
        # so XH bondconstants are already calculated by QM
        # Also one can manually change the XY bond lengths
        self.write_eq_distances_to_file(\
            self.qms_edge)

        
        #get target charge of each qm-region
        self.classical_target_charge_sums = \
            self.get_classical_target_charge_sums\
            (self.mm_calculator.topology_filename, self.qms)


        #get a list of link H atoms
        self.link_atoms = self.get_link_atoms(\
            self.qms_edge, self.mms_edge,\
                self.force_constants,\
                self.equilibrium_distances_xh, \
                self.equilibrium_distances_xy)

        self.qmsystems = self.define_QM_clusters_in_vacuum(system_tmp)

        for iqm, qm in enumerate(self.qmsystems):
            write('test-qm-'+str(iqm)+'.xyz', qm)
        

        #attach calculators to qm regions
        for iqm, qm in enumerate(self.qmsystems):
            self.qmsystems[iqm].set_calculator(self.qm_calculators[iqm])

        #attach calculators to the mm region (the whole system)
        self.mm_system = system_tmp
        self.mm_system.set_calculator(self.mm_calculator)

        #initialize total energy and forces of qm regions
        #and the mm energy
        self.qm_energies = []
        self.qm_forces = []
        self.qm_charges = []
        self.sum_qm_charge = []
        for iqm, qm in enumerate(self.qmsystems):
            self.qm_energies.append(0.0)
            self.qm_forces.append(None)
            self.qm_charges.append(None)
            self.sum_qm_charge.append(None)
        self.mm_energy = None
        #set initial zero forces
        self.forces = np.zeros((len(self.positions), 3))
        self.charges = np.zeros((len(self.positions), 1))

        try:
            os.remove(self.mm_calculator.topology_filename+'.orig')
        except:
            pass
        print('%s' % str(self.mm_calculator.topology_filename))
        os.system('cp ' + self.mm_calculator.topology_filename + ' ' +\
                      self.mm_calculator.topology_filename + '.orig')

        #remove some classical bonded interaction in the topology file
        # this need to be done only once, because the bond topology
        # is unchanged during a QM/MM run
        #(QM charges can be updated in the topology, however)

        # the original topology is generated when calling Gromacs(
        # in the main script setting up QM, MM and minimization

        if (self.mm_calculator.name == 'Gromacs'):
            self.kill_top_lines_containing_only_qm_atoms\
                (self.mm_calculator.topology_filename, self.qms, \
                     self.mm_calculator.topology_filename)
        else:
            print('Only Gromacs MM-calculator implemented in ASE-QM/MM')
            sys.exit()
        #exclude qm-qm non-bonded interactions in MM-gromacs
        self.add_exclusions()
        #generate input file for gromacs run
        self.mm_calculator.generate_gromacs_run_file()

        ######### end of Init #####################################

    def get_forces(self, atoms):
        """get forces acting on all atoms except link atoms """
        self.update(atoms)
        return self.forces

    def get_potential_energy(self, atoms):
        """ get the total energy of the MM and QM system(s) """
        self.update(atoms)
        return self.energy

    def update(self, atoms):
        """Updates and does a check to see if a calculation is required"""



        if self.calculation_required(atoms):
            # performs an update of the atoms and qm systems
            self.atoms = atoms.copy()
            self.positions = atoms.get_positions()
            self.mm_system = atoms.copy()

            #get the positions of link H atoms
            self.link_atoms = self.get_link_atoms(\
                self.qms_edge, self.mms_edge,\
                    self.force_constants,\
                    self.equilibrium_distances_xh, \
                    self.equilibrium_distances_xy)

            #get QM systens
            self.qmsystems = self.define_QM_clusters_in_vacuum(\
                self.atoms)
            self.calculate(atoms)

    def calculation_required(self, atoms):
        """Checks if a calculation is required"""
        if ((self.positions is None) or
            (self.atoms != atoms) or
            (self.energy is None)):
            return True
        return False


    def calculate_mm(self):
        """ Calculating mm energies and forces """
        import os
        mm = self.atoms
        mm.set_calculator(self.mm_calculator)
        if (self.mm_calculator.name == 'Gromacs'):
            try:
                os.remove(self.mm_calculator.label+'.log')
            except:
                pass
        self.mm_calculator.update(mm)
        #self.mm_calculator.run()
        #self.mm_calculator.calculate(atoms=mm, properties=['energy', 'forces'])
                                     
        self.mm_energy = 0
        
        self.mm_energy += mm.get_potential_energy()
        self.forces += mm.get_forces()

    def calculate_qms(self):
        """ QM calculations on all qm systems are carried out """
        for iqm, qm in enumerate(self.qmsystems):
            qm.set_calculator(self.qm_calculators[iqm])
            self.qm_energies[iqm] = qm.get_potential_energy()
            self.qm_forces[iqm] = np.zeros((len(qm), 3))
            self.qm_forces[iqm] = qm.get_forces()
            (self.sum_qm_charge[iqm], self.qm_charges[iqm]) = \
                self.get_qm_charges(iqm,
                                    number_of_link_atoms =\
                                        len(self.qms_edge[iqm]))
            if (len(self.qms[iqm]) != len(self.qm_charges[iqm])):
                print('Problem in reading charges')
                print('len(self.qms[iqm]) %s' % str(len(self.qms[iqm])))
                print('len(self.qm_charges[iqm]) %s'
                      % str(len(self.qm_charges[iqm])))
                print('Check the output of QM program')
                print('iqm, qm %s' % str(iqm)+ ' '+ str(qm))
                print('self.qm_charges[iqm] %s' % str(self.qm_charges[iqm]))
                sys.exit()

    def calculate_single_qm(self, myqm, mycalculator):
        """ Calculate the qm energy of a single qm region
        (for X-H bond length calculations)
        """
        myqm.set_calculator(mycalculator)
        return myqm.get_potential_energy()

    def run(self, atoms):
        """Runs QMs and MM"""

        self.forces = np.zeros((len(atoms), 3))

        self.calculate_qms()

        # update QM charges to MM topology file
        self.set_qm_charges_to_mm_topology()

        #generate gromacs run file (.tpr) base on new topology
        self.mm_calculator.generate_gromacs_run_file()
        self.calculate_mm()
                        

    def calculate(self, atoms):
        """gets all energies and forces (qm, mm, qm-mm and corrections)"""

        self.run(atoms)

        self.energy = sum(self.qm_energies)+self.mm_energy
        
        #map the forces of QM systems to all atoms
        #loop over qm regions
        for qm, qm_force in zip(self.qms, self.qm_forces):
            #loop over qm atoms in a qm region
            #set forces to the all-atom set (the all atom set does not
            # have link atoms)
            for iqm_atom, qm_atom in enumerate(qm):
                self.forces[qm_atom] = self.forces[qm_atom] + \
                    qm_force[iqm_atom]
        self.get_link_atom_forces(action = 'QM')


    def get_link_atoms(self, qm_links, mm_links, \
                        force_constants,\
                        equilibrium_distances_xh, equilibrium_distances_xy):
        """
        QM atoms can be bonded to MM atoms. In this case one sets
        an extra H atom (a link atom).

        The positions of the all link H atoms in all qm regions are
        set along QM-MM and bond with length defined by:

        J. Chem. Theory Comput 2011, 7, 761-777, Eq 1
        r_XH = r_XY_current*(r_XH_from_qm_calculation /r_XY_from_forceField)

        """
        import math
        from ase import Atom

        link_hs = []

        for i_qm_region, (qm0, mm0) in enumerate (zip( qm_links, mm_links)):
            for i_link_atom, (qmatom, mmatom) in enumerate (zip(qm0, mm0)):
                dx = (self.positions[mmatom, 0] - self.positions[qmatom, 0])
                dy = (self.positions[mmatom, 1] - self.positions[qmatom, 1])
                dz = (self.positions[mmatom, 2] - self.positions[qmatom, 2])
                d = math.sqrt(dx* dx+ dy* dy+ dz* dz)
                
                unit_x = dx/ d
                unit_y = dy/ d
                unit_z = dz/ d
                xh_bond_length = \
                    d*\
                    self.equilibrium_distances_xh[i_qm_region][i_link_atom]/\
                    self.equilibrium_distances_xy[i_qm_region][i_link_atom]
            
                posh_x = self.positions[qmatom, 0] + unit_x* xh_bond_length
                posh_y = self.positions[qmatom, 1] + unit_y* xh_bond_length
                posh_z = self.positions[qmatom, 2] + unit_z* xh_bond_length
                tmp_link_h = (Atom('H', position=(posh_x, posh_y, posh_z)))
                link_h = LinkAtom(atom=tmp_link_h, \
                                      qm_region_index = i_qm_region,\
                                      link_atom_index = i_link_atom)
                link_h.set_link_atom_qm_neighbor(qmatom)
                link_h.set_link_atom_mm_neighbor(mmatom)
                link_h.set_force_constant(\
                    force_constants[i_qm_region][i_link_atom])
                link_h.set_equilibrium_distance_xh(equilibrium_distances_xh\
                                      [i_qm_region][i_link_atom])
                link_h.set_equilibrium_distance_xy(equilibrium_distances_xy\
                                      [i_qm_region][i_link_atom])


                link_hs.append(link_h)
        return (link_hs)



    def get_link_atom_forces(self, action):
        """ Add forces due to link atom to QM atom
        and to MM atom next to each link atom.

        Top Curr Chem (2007) 268: 173-290
        QM/MM Methods for Biological Systems
        Hans Martin Senn and Walter Thiel

        Eqs. 10(p192), 12(p193), 16a, 16b(p 194)

        """

        for link_atom in self.link_atoms:
            i_qm_atom = link_atom.qm_neighbor
            i_mm_atom = link_atom.mm_neighbor
            i_qm_region = link_atom.qm_region_index
            link_atom_index_in_qm = link_atom.get_link_atom_index_in_qm()

            if (action == 'QM'):
                force_of_h = self.qm_forces[i_qm_region][link_atom_index_in_qm]
            elif (action == 'MM'):
                force_of_h = link_atom.mm_force
            else:
                print('not implemented in get_link_atom_forces')
                sys.exit()

            g = link_atom.equilibrium_distance_xh/\
                link_atom.equilibrium_distance_xy
            
            self.forces[i_mm_atom, 0]  = self.forces[i_mm_atom, 0] +\
                force_of_h[0] * g
            self.forces[i_mm_atom, 1]  = self.forces[i_mm_atom, 1] +\
                force_of_h[1] * g
            self.forces[i_mm_atom, 2]  = self.forces[i_mm_atom, 2] +\
                force_of_h[2] * g

            self.forces[i_qm_atom, 0]  = self.forces[i_qm_atom, 0] +\
                force_of_h[0] * (1.0 - g)
            self.forces[i_qm_atom, 1]  = self.forces[i_qm_atom, 1] +\
                force_of_h[1] * (1.0 - g)
            self.forces[i_qm_atom, 2]  = self.forces[i_qm_atom, 2] +\
                force_of_h[2] * (1.0 - g)





    def add_energy_exclusion_group(self, indexfilename='index.ndx'):
        """
        Add energy exclusions for MM calculations.
        This is the way to block non-bonded MM (coulomb&vdW)
        interactions within a single QM region.
        """
        
        infile = open(indexfilename,'r')
        lines = infile.readlines()
        infile.close()
        
        qm_region_names = []

        for line in lines:
            if (('QM' in line) or ('Qm' in line) or ('qm' in line)):
                qm_region_names.append(line.split()[1])

        infile = open(self.mm_calculator.label+'.mdp','r')
        lines = infile.readlines()
        infile.close()
        outfile = open(self.mm_calculator.label+'.mdp','w')
        for line1 in lines:
            outfile.write(line1)
        outfile.write(';qm regions should not MM-interact with themselves \n')
        outfile.write(';but separate qm regions MM-interact with each other \n')

        outfile.write('energygrps = ')
        for name in qm_region_names:
            outfile.write(name + ' ')
        outfile.write('\n')
        outfile.write('energygrp_excl = ')
        for name in qm_region_names:
            outfile.write(name + ' ' + name + ' ')
        outfile.write('\n')
        outfile.close()

        return

    def add_exclusions(self):
        """
        Add energy exclusions for MM calculations.
        This is the way to block non-bonded MM (coulomb&vdW)
        interactions within a single QM region.
        """
        

        infile = open(self.mm_calculator.topology_filename,'r')
        lines = infile.readlines()
        infile.close()
        outfile = open(self.mm_calculator.topology_filename,'w')
        for line in lines:
            if '[ angle' in line:
                outfile.write('\n')
                outfile.write('[ exclusions ] \n')
                outfile.write(\
                    '; qm regions should not MM-interact with themselves \n')
                outfile.write(\
                    '; but separate qm regions MM-interact with each other \n')

                for qm_region in self.qms:
                    for qm_atom1 in qm_region:
                        outfile.write(str(qm_atom1 + 1) + ' ')
                        for qm_atom2 in qm_region:
                            if qm_atom1 != qm_atom2:
                                outfile.write(str(qm_atom2 + 1) + ' ')
                        outfile.write('\n')
                outfile.write('\n')
            outfile.write(line)



        outfile.close()

        return

    def get_qm_charges(self, i_current_qm, calculator='Aims',
                       number_of_link_atoms = 0):
        """
        Get partial charges on QM atoms.
        The charges at link atoms are not returned.
        """
        if calculator == 'Aims':
            infile = open('aims'+str(i_current_qm)+'.out','r')
            lines = infile.readlines()
            infile.close()
            qm_charges = []
            for line in lines:
                if ('Hirshfeld charge      ' in line):
                    qm_charges.append(float(line.split()[4]))
            sum_qm_charges = sum(qm_charges)
            #delete charges of link atoms
            if (number_of_link_atoms > 0):
                del qm_charges[-number_of_link_atoms:]
                    
        return sum_qm_charges, qm_charges

    def get_topology_lines(self, lines):
        """ Get lines including charges of atoms (ok_lines)
        also comments in these lines (comment_lines)
        and lines before and after these lines
        (lines_before and lines_after)
        """
        lines_before = []
        lines_change = []
        lines_after = []
        do_lines_before = True
        do_lines_change = False
        for line in lines:
            if (' bonds ') in line:
                do_lines_change = False
            if do_lines_before:
                lines_before.append(line)
            elif do_lines_change:
                lines_change.append(line)
            else:
                lines_after.append(line)
            
            if (' atoms ') in line:
                do_lines_before = False
                do_lines_change = True

        #kill comments and empty lines,
        #get the charge in the topology file
        comment_lines = []
        lines_ok = []
        for iline in range(len(lines_change)):
            if lines_change[iline].startswith(';'):
                comment_lines.append(lines_change[iline])
            elif not lines_change[iline].strip():
                pass
            else:
                try:
                    #new charge = float(lines_change[iline].split()[6])
                    #new charge_orig = charge_orig + charge
                    #top_charge.append(charge)
                    lines_ok.append(lines_change[iline])
                except:
                    print('error in reading gromacs topology')
                    print('line is')
                    print('%s' % lines_change[iline])
                    sys.exit()
        return lines_before, comment_lines, lines_ok, lines_after


    def set_qm_charges_to_mm_topology(self):
        """ Set qm charges to qm atoms of MM topology based on
        a QM calculation.
        1) The charges of link atoms are neglected.
        2) The charge of a qm atom next to the link atom is set to be the
        same value as in the original topology file. (trying to
        avoid the artificial polarization due to qmAtom-linkH).
        3) the total charge of the system (all QM and MM atoms) should be
        the same as in the original classical system. Therefore, all the
        QM atoms will gain/loose an equal amount of charge in the MM topology
        file.
        """

        infile = open(self.mm_calculator.topology_filename,'r')
        lines = infile.readlines()
        infile.close()

        (lines_before, comment_lines, lines_ok, lines_after) = \
            self.get_topology_lines(lines)

        #check that the atom numering is ok
        for iline in range(len(lines_ok)):
            atom_nr = iline + 1
            if int(lines_ok[iline].split()[0]) != atom_nr:
                print('2: error in reading gromacs topology')
                print('line is')
                print('%s' % lines_ok[iline])
                sys.exit()
        
        # get the total charge of non-link H atoms in the current qm system
        # The charges of edge atoms and their H neighbors
        # are taken from topology
        # (they are unchanged, it is not from QM calculations)
        for iqm, qm in enumerate(self.qms):
            charges = self.qm_charges[iqm]
            charges_ok = charges
            qm_charge_no_link_edge_mm = 0.0
            n_qm_charge_atoms = 0

            for qm_atom, charge in zip(qm, charges):
                if qm_atom not in self.constant_charge_qms:
                    qm_charge_no_link_edge_mm = \
                    qm_charge_no_link_edge_mm + charge
                    n_qm_charge_atoms = n_qm_charge_atoms + 1

            # correct the total charge to be equal the original one
            # in the topology file by
            # adding/ subtracting missing/extra charge on
            # non-edge and non-single neighbor next neib QM atoms
            change_charge = \
                ( self.classical_target_charge_sums[iqm] - \
                     qm_charge_no_link_edge_mm)/\
                     float(n_qm_charge_atoms)
            for iqmatom, qmatom in enumerate(qm):
                if qmatom not in self.constant_charge_qms:
                    charges_ok[iqmatom] = charges[iqmatom] + change_charge
            # set qm charges to the lines of gromacs topology file
            for iqmatom, qmatom in enumerate(qm):
                if qmatom not in self.constant_charge_qms:
                    lines_ok[qmatom] = \
                        lines_ok[qmatom][0:45]\
                        +str(round((charges_ok[iqmatom]),5)).rjust(11)+\
                        lines_ok[qmatom][56:70]
        # write out the new topology file
        sum_charge = 0.0
        for iline in range(len(lines_ok)):
            sum_charge = sum_charge + float(lines_ok[iline][46:56])
            comment = '; qtot '+str(round(sum_charge,4))+'\n'.ljust(12)
        
        outfile = open(self.mm_calculator.topology_filename, 'w')
        for line in lines_before:
            outfile.write(line)
        for line in comment_lines:
            outfile.write(line)
        sum_charge = 0.0
        for line in lines_ok:
            sum_charge = sum_charge + float(line[46:56])
            comment = '; qtot '+str(round(sum_charge,4)).ljust(11)+'\n'
            outfile.write(line[0:70]+comment)
        outfile.write('\n')
        for line in lines_after:
            outfile.write(line)
        outfile.close()


#------------------------------------------------------------------
#------Below the stuff needed for initializing the QM/MM system ---
#------Setting up link atoms, defining QM and MM regions ----------
#------------------------------------------------------------------




    def get_edge_qm_and_mm_atoms(self, qms, system):
        """  Get neighbors of QM atoms (MM-link-atoms) that are not in QM
        (there may be many QM regions)
        edge-QM atom can NOT be neighbored by H atom(s)
        also get edge-QM atoms
        """


        masses = system.get_masses()

        mms1 = []
        qms1 = []
        setmms1 = set([])
        setqms1 = set([])
        for qm in qms:

            link_mm_atoms = []
            link_qm_atoms = []
            for qm_atom in qm:
                indices, offsets = self.neighbor_list.get_neighbors(qm_atom)
                for neib_atom in indices:
                    if neib_atom not in qm:
                        link_mm_atoms.append(neib_atom)
            #take unique atoms of flattened list
            link_mm_atoms = list(set(link_mm_atoms))
            # Kill MM atoms that are H atoms in the neighborlist
            oklink_mm_atoms = []
            for index in link_mm_atoms:
                if masses[index] > 1.5:
                    oklink_mm_atoms.append(index)
                else:
                    print('WARNING:')
                    print('qm system cannot be bond to H atoms')
                    print('problem atom index is (numbering from 1): %s' %
                          str(index+1))
                    print('if this is water H you should consider including it')
                    print('in QM')
                    #sys.exit()


            #get indexes of QM edge atoms,
            # one qm atom can be more then one time an edge atom
            # (then this QM atom will have more than one link atoms)
            for link_mm_atom in oklink_mm_atoms:
                indices, offsets = \
                    self.neighbor_list.get_neighbors(link_mm_atom)
                for neib_atom in indices:
                    if neib_atom in qm:
                        link_qm_atoms.append(neib_atom)


            mms1.append(oklink_mm_atoms)
            qms1.append(link_qm_atoms)
            setmms1 |= set(oklink_mm_atoms)
            setqms1 |= set(link_qm_atoms)
        return mms1, qms1, setmms1, setqms1

    def get_next_neighbors(self, atom_indexes, prohibited_set):
        """  Get neighbors of all atoms in 'atom_indexes'
        that are not in 'prohibited_set'.

        'atom_indexes' is a list of list in which atom indexes belonging
        of each QM region is a separate list, that is
        [[QM1 atom_indexes], [QM2 atom_indexes], ...]
        """

        list_neibs = []
        set_list_neibs = set([])
        for current_atoms in atom_indexes:
            neibs = []
            set_current_atoms = set(current_atoms)
            for current_atom in current_atoms:
                indices, offsets = \
                    self.neighbor_list.get_neighbors(current_atom)
                setneib = set(indices)
                neibs += list(setneib - set_current_atoms-prohibited_set)
            list_neibs.append(neibs)
            set_list_neibs |= set(neibs)
        return list_neibs, set_list_neibs

    def get_constant_charge_qms(self, set_qms_edge, set_second_qms):
        """ get indices of all qm atoms whose charge in MM
        calculations is taken from the original MM-topology
        (not from the QM calculation). These atoms are edge QM atoms
        and their neighbors in QM which have only one neighbor.
        At least C(edge-qm)-H(second-edge-qm) and C(edge-qm)=O(second-edge-qm)
        """

        set_charge_exclusion = set_qms_edge

        for second_qms in set_second_qms:
            indices, offsets = self.neighbor_list.get_neighbors(second_qms)
            if len(indices)== 1:
                set_charge_exclusion.add(second_qms)
        return set_charge_exclusion


    def get_eq_distances_xy(\
        self, topfilename = 'gromos.top', force_field = 'oplsaa'):
        """
        The link atom is positioned as in
        J. Chem. Theory Comput 2011, 7, 761-777, Eq 1

        For this purpose we need the equilibrium length of each
        QM-MM covalent bond. Those are obtained here from the
        files of the force field.

        """

        import os

        print('in get_eq_distances_xy, topfilename=')
        print('%s' % topfilename)
        for qm in self.qms_edge:
            equilibrium_distance_xy = []
            for iqm in qm:
                equilibrium_distance_xy.append(0.0)
            self.equilibrium_distances_xy.append(equilibrium_distance_xy)

        #get the version of the topology file where one sees the bond
        # force constants (file is named as gromacs.top.dump)
        try:
            os.remove(self.mm_calculator.label+'.tpr.dump')
        except OSError:
            pass

        os.system('gmxdump -s '+ self.mm_calculator.label\
                      +'.tpr > ' + \
                  self.mm_calculator.label+ \
                      '.tpr.dump 2>/dev/null')
        if 'GMXDATA' in os.environ:
            gromacs_home = os.environ['GMXDATA'].split(':')[0]
        else:
            gromacs_home = '/usr/local/gromacs/share/gromacs/'

        #read the bonded force constants of this force field in order to
        #get an estimate for X-Y bond constant
        linesff = open(gromacs_home+ '/top/'+ force_field+ \
                           '.ff/ffbonded.itp', 'r').readlines()
        oklinesff = []
        start = False

        for line in linesff:
            if 'bondtypes' in line:
                start = True
            elif '[' in line:
                break
            if start and (line.strip()):
                oklinesff.append(line)

        #lines for getting oplsaa atom dual-types
        if 'opls' in force_field:
            lines_for_dual_types = open(gromacs_home+ '/top/'+ force_field+ \
                                         '.ff/ffnonbonded.itp', 'r').readlines()

        #read the types of interaction for bond stretching
        lines_tpr = open(self.mm_calculator.label+\
                             '.tpr.dump', 'r').readlines()

        #read the topology file to get QM atom type
        lines_top = open(topfilename, 'r').readlines()
        oklines_top = []
        start = False
        for line in lines_top:
            if start and ('[' in line):
                break
            if start:
                if (not line.startswith(';')) or (not line.strip()):
                    oklines_top.append(line)
            if '[ atoms' in line:
                start = True


        #get force constant and bond eq distance for all QM-MM bonds
        #
        ok_equilibrium_distances_xy = []
        ok_qmatom_types = []
        ok_mmatom_types = []
        for qm0, mm0, eqsxy in zip(
            self.qms_edge, self.mms_edge, \
                self.equilibrium_distances_xy):
            ok_eqxy = []
            ok_qmatom_type = []
            ok_mmatom_type = []
            for qmatom, mmatom, eqxy in \
                    zip(qm0, mm0, eqsxy):
                #find qm-mm bond in topology file (indexes from 0)
                # get the index for interaction
                interaction = 'empty'
                for line in lines_tpr:
                    if (' type' in line) and ('BONDS' in line):
                        if (qmatom == int(line.split()[3])) and \
                                (mmatom == int(line.split()[4])):
                            interaction = line.split()[1].lstrip('type=')
                            break
                        if (qmatom == int(line.split()[4])) and \
                                (mmatom == int(line.split()[3])):
                            interaction = line.split()[1].lstrip('type=')
                            break
                if interaction == 'empty':
                    print('QM-MM bond not found in topology')
                    print('atoms are: QM, MM: (from 1 indexing) %s' %
                          str(qmatom + 1) + '  ' + str(mmatom + 1))
                    sys.exit()
                for line in lines_tpr:
                    if ('functype['+interaction+']=BONDS') in line:
                        r_xy0 = float(line.split()[2].rstrip(','))
                #get type of the QM atom
                qmatom_type = 'empty'
                for line in oklines_top:
                    if (int(line.split()[0] ) == qmatom+ 1):
                        qmatom_type = line.split()[1]
                        #oplsaa atom type has a double name,
                        #the other one is used in file ffbonded.itp
                        break
                if (qmatom_type == 'empty'):
                    print('problem in QM atom type')
                    sys.exit()

                if 'opls' in force_field:
                    found = False
                    for line in lines_for_dual_types:
                        if (qmatom_type == line.split()[0]):
                            qmatom_type = line.split()[1]
                            found = True
                            break
                    if not found:
                        print('problem in QM atom type')
                        print('with OPLSAA force field dual atom types')
                        sys.exit()
                #get type of the true link-MM atom
                mmatom_type = 'empty'
                for line in oklines_top:
                    if (int(line.split()[0] ) == mmatom+ 1):
                        mmatom_type = line.split()[1]
                        #oplsaa atom type has a double name,
                        #the other one is used in file ffbonded.itp
                        break
                if (mmatom_type == 'empty'):
                    print('problem in MM atom type')
                    sys.exit()

                if 'opls' in force_field:
                    found = False
                    for line in lines_for_dual_types:
                        if (mmatom_type == line.split()[0]):
                            mmatom_type = line.split()[1]
                            found = True
                            break
                    if not found:
                        print('problem in MM atom type')
                        print('with OPLSAA force field dual atom types')
                        sys.exit()

                ok_qmatom_type.append(qmatom_type)
                ok_mmatom_type.append(mmatom_type)

                if (eqxy != 0.0):
                    #use eq constant given by the user
                    ok_eqxy.append(eqxy)
                else:
                    ok_eqxy.append(r_xy0)
            ok_equilibrium_distances_xy.append(ok_eqxy)
            ok_qmatom_types.append(ok_qmatom_type)
            ok_mmatom_types.append(ok_mmatom_type)
        outfile = open('qm-mm-linkAtomsInfo.txt','w')
        outfile.write(\
            '=======================================================\n')
        outfile.write('Information about QM-MM boundary(ies) \n')
        outfile.write(\
            'Created using the Atomic Simulation Environment (ASE) \n')
        outfile.write(\
            '=======================================================\n')
        qmregion_count = 0

        # ADD qm-mm-linkAtomsInfo.txt
        for qm, mm, eqs_xy, eqs_xh, qmtypes, mmtypes in zip\
                (self.qms_edge, self.mms_edge, ok_equilibrium_distances_xy,\
                     self.equilibrium_distances_xh,\
                     ok_qmatom_types, ok_mmatom_types):
            outfile.write(\
                '=======================================================\n')
            qmregion_count = qmregion_count+ 1
            outfile.write('Parameters related to QM region number '+\
                             str(qmregion_count)+'\n')
            for qmatom, mmatom,  eq_xy, eq_xh, qmtype, mmtype in zip\
                    (qm, mm, eqs_xy, eqs_xh,\
                    qmtypes, mmtypes):
                outfile.write('qm-link-atom-index (from 1): '+str(qmatom)+'\n')
                outfile.write('qm-link-atom-type: '+str(qmtype)+'\n')
                outfile.write('mm-link-atom-index (from 1): '+str(mmatom)+'\n')
                outfile.write('mm-link-atom-type: '+str(mmtype)+'\n')
                outfile.write('qm-mm(notH)-equilibrium-distance: '\
                                  +str(eq_xy)+' nm\n')
                outfile.write('qm-H-equilibrium-distance(calculated by QM): '\
                                  +str(eq_xh)+' nm\n')
        outfile.close()
        self.equilibrium_distances_xy = ok_equilibrium_distances_xy
        self.qmatom_types = ok_qmatom_types
        self.mmatom_types = ok_mmatom_types

        return

    def write_eq_distances_to_file(
        self,
        qm_links, filename='linkDATAout.txt'):
        """
        Write classical bond equilibrium lengths
        for XY (X in QM, Y in MM)
        Write QM calculated XH(link atom) bond length (X in QM, H link atom)
        """

        outfile = open(filename, 'w')
        for iqm_region, qmlink in enumerate (qm_links):
            for ilink, dummy in enumerate (qmlink):
                data = self.equilibrium_distances_xy[iqm_region][ilink]
                outfile.write(str(data)+' ')
                outfile.write('\n')
                data = self.equilibrium_distances_xh[iqm_region][ilink]
                outfile.write(str(data)+' ')
                outfile.write('\n')
                data = self.force_constants[iqm_region][ilink]
                outfile.write(str(data)+' ')
                outfile.write('\n')
                data = self.qmatom_types[iqm_region][ilink]
                outfile.write(str(data)+' ')
                outfile.write('\n')
                data = self.mmatom_types[iqm_region][ilink]
                outfile.write(str(data)+' ')
                outfile.write('\n')
        outfile.close()
        return


    def read_eq_distances_from_file(self, filename='linkDATAin.txt'):
        """
        Read classical bond equilibrium lengths
        for XY (X in QM, Y in MM) or XH (X in QM, H link atom)
        """

        myfile = open(filename, 'r')
        self.equilibrium_distances_xy = []
        self.equilibrium_distances_xh = []
        self.force_constants = []
        self.qmatom_types = []
        self.mmatom_types = []
        print('Reading X-H and other data from file: %s' % filename)

        for qm in self.qms_edge:
            equilibrium_distance_xy = []
            equilibrium_distance_xh = []
            force_constant = []
            qmatom_type = []
            mmatom_type = []
            for iqm, dum in enumerate(qm):
                line = myfile.readline()
                equilibrium_distance_xy.append(float(line.split()[0]))
                line = myfile.readline()
                equilibrium_distance_xh.append(float(line.split()[0]))
                line = myfile.readline()
                force_constant.append(float(line.split()[0]))
                line = myfile.readline()
                qmatom_type.append(line.split()[0])
                line = myfile.readline()
                mmatom_type.append(line.split()[0])

            self.equilibrium_distances_xy.append(equilibrium_distance_xy)
            self.equilibrium_distances_xh.append(equilibrium_distance_xh)
            self.force_constants.append(force_constant)
            self.qmatom_types.append(qmatom_type)
            self.mmatom_types.append(mmatom_type)
        myfile.close()
        return

    def get_eq_qm_atom_link_h_distances(self, system_tmp):
        """ get equilibrium QMatom-linkH distances
        for all linkH:s
        by  QM  """


        #import matplotlib
        #matplotlib.use('Agg')
        #import matplotlib.pyplot as plt

        from scipy.optimize import fmin

        def qm_bond_energy_function(x, system_tmp, i_qm_region):
            """ get the qm energy of a single qm system with a given
            edge-qm-atom---link-h-atom distances of that qm region
            The qm region is i_qm_region, all
            edge-qm-atom---link-h-atom distance in this qm_region are
            optimized simultaneously
            """

            for index_x, current_x in enumerate(x):
                self.equilibrium_distances_xh\
                    [i_qm_region][index_x] = current_x
            print('current X-H bond lengths [nm]')
            print('%s' % str(x))
            self.link_atoms = self.get_link_atoms(\
                self.qms_edge, self.mms_edge,\
                    self.force_constants,\
                    self.equilibrium_distances_xh, \
                    self.equilibrium_distances_xy)
            self.qmsystems = \
                self.define_QM_clusters_in_vacuum(system_tmp)
            
            #try:
            single_qm_energy = self.calculate_single_qm(\
                self.qmsystems[i_qm_region],\
                    self.qm_calculators[i_qm_region])
            #except RuntimeError:
            #    single_qm_energy = BIG_VALUE
            return single_qm_energy

        print('=====================================================')
        print('Calculating X-H bond lengths and bond force constants')
        print('by QM in one shot for each QM region.')
        print('In later calculations you can: ')
        print('cp linkDATAout.txt linkDATAin.txt')
        print("and set link_info = 'byFILE'")
        print('=====================================================')


        self.equilibrium_distances_xh = []
        self.force_constants = []
        for qm_edges in self.qms_edge:
            force_constants = []
            equilibrium_distances_xh = []
            for qm_edge in qm_edges:
                force_constants.append(0.0)
                equilibrium_distances_xh.append(0.11)
            self.force_constants.append(force_constants)
            self.equilibrium_distances_xh.append(equilibrium_distances_xh)

        #loop over qm regions. To get optimal simultaneous
        # edgeQMatom-linkH distance(s) in [nm] in that qm region
        for i_qm_region in range(len(self.qms_edge)):
            print('NOW running : ')
            print('QM region for optimising edge-linkH distances %s' %
                  str(i_qm_region))
            x = self.equilibrium_distances_xh[i_qm_region][:]
            xopt = fmin(qm_bond_energy_function, \
                            x,\
                            args=(system_tmp, i_qm_region),\
                            xtol=0.0001, ftol=0.0001)
            for index_xopt, current_xopt in enumerate(xopt):
                self.equilibrium_distances_xh\
                    [i_qm_region][index_xopt] = current_xopt
                print('i_qm_region, i_link_atom, optimal X-H bond[nm] %s' \
                      % (str(i_qm_region) + ' ' + str(index_xopt) \
                         + ' ' + str(current_xopt)))

    def define_QM_clusters_in_vacuum(self, system):
        """ Returns Each QM system as an Atoms object
        We get a list of these Atoms objects
        (in case we have many QM regions).
        """
        from ase import Atoms

        qmsystems = []
        for qm0 in self.qms:
            tmp_system = Atoms()
            for qmatom in qm0:
                tmp_system += system[qmatom]
            qmsystems.append(tmp_system)
        for link_atom in self.link_atoms:
            tmp_atom = link_atom.get_link_atom()
            qm_region = link_atom.get_link_atom_qm_region_index()
            link_atom_index_in_qm = len(qmsystems[qm_region])
            qmsystems[qm_region].append(tmp_atom)
            link_atom.set_link_atom_index_in_qm(link_atom_index_in_qm)
        return qmsystems



    def kill_top_lines_containing_only_qm_atoms(self, \
                                              intopfilename, \
                                              qms, outtopfilename):
        """
            Delete all lines in the topology file that contain only qm atoms
            in bonded sections
            (bonds, angles or dihedrals)
            and in pairs section (1-4 interactions)
            """

        # get an index of all qm atoms in all qm regions
        qm = set()
        for qm_tmp in qms:
            qm = qm.union(set(qm_tmp))

        infile = open(intopfilename,'r')
        lines = infile.readlines()
        infile.close()
        outfile = sys.stdout
        oklines = []

        accept = True
        check = ''
        for line in lines:
            if (('[ bonds' in line)):
                oklines.append(line)
                accept = False
                check = 'bond'
            elif (('[ angles' in line)):
                oklines.append(line)
                accept = False
                check = 'angle'
            elif (('[ dihedrals' in line)):
                oklines.append(line)
                accept = False
                check = 'dihedral'
            elif (('[ pairs' in line)):
                oklines.append(line)
                accept = False
                check = 'pair'
            elif ('[' in line):
                oklines.append(line)
                accept = True
                check = ''
            elif line in ['\n']:
                oklines.append(line)
                accept = True
                check = ''
            elif accept:
                oklines.append(line)
            else:
                indexes = [int(float(s)-1.0) \
                               for s in line.split() if s.isdigit()]
                indexes1 = [int(s) for s in line.split() if s.isdigit()]
                if indexes == []:# this takes comment line
                                 #after bond, angle, dihedral
                    oklines.append(line)
                elif check == 'bond':
                    bondedatoms = set(indexes[0:2])
                    #set empty bond intereaction for qm-qm bonds (type 5)
                    #(this way LJ and electrostatics is not messed up)
                    if (bondedatoms.issubset(qm)):
                        newline = str(indexes1[0]).rjust(8)+\
                            str(indexes1[1]).rjust(8)+\
                            ('5').rjust(8) + '\n'
                        oklines.append(newline)
                    else:
                        oklines.append(line)
                elif check == 'angle':
                    bondedatoms = set(indexes[0:3])
                    if (bondedatoms.issubset(qm)):
                        pass
                    else:
                        oklines.append(line)
                elif check == 'dihedral':
                    bondedatoms = set(indexes[0:4])
                    if (bondedatoms.issubset(qm)):
                        pass
                    else:
                        oklines.append(line)
                elif check == 'pair':
                    bondedatoms = set(indexes[0:2])
                    if (bondedatoms.issubset(qm)):
                        pass
                    else:
                        oklines.append(line)
        outfile = open(outtopfilename,'w')
        for line in oklines:
            outfile.write(line)
        outfile.close()

        return


    def get_classical_target_charge_sums(self, intopfilename, qms):
        """ get sum of MM charges of the charged changed by QM
        these are qm atoms that are not link-atoms or edge-qm atoms

        xxx this has a problem:
        Water is in .itp files, not in topology...
        """
        infile = open(intopfilename,'r')
        lines = infile.readlines()
        infile.close()

        (lines_before, comment_lines, ok_lines, lines_after) = \
            self.get_topology_lines(lines)


        classical_target_charge_sums = []
        for iqm, qm in enumerate(qms):
            classical_target_charge_sum = 0.0
            for line in ok_lines:
                atom_index = int(line.split()[0])-1
                if (atom_index in qm) and \
                        (not(atom_index in self.constant_charge_qms)):
                    classical_target_charge_sum = \
                        classical_target_charge_sum + \
                        float(line.split()[6])
            classical_target_charge_sums.\
                append(classical_target_charge_sum)
        return classical_target_charge_sums



import json
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
from ase.collections import dcdft
from ase.eos import birchmurnaghan
from ase.io import read
from ase.units import kJ
from ase.utils.deltacodesdft import delta

# Read EMT data:
data = json.loads(Path('fit.json').read_text())
# Insert values from experiment and WIEN2k:
for symbol in data:
    dcdft_dct = dcdft.data[symbol]
    dcdft_dct['exp_B'] *= 1e-24 * kJ
    dcdft_dct['wien2k_B'] *= 1e-24 * kJ
    data[symbol].update(dcdft_dct)

for name in ['volume', 'B', 'Bp']:
    with open(name + '.csv', 'w') as f:
        print('# symbol, emt, exp, wien2k', file=f)
        for symbol, dct in data.items():
            values = [dct[code + '_' + name]
                      for code in ['emt', 'exp', 'wien2k']]
            if name == 'B':
                values = [val * 1e24 / kJ for val in values]
            print('{},'.format(symbol),
                  ', '.join('{:.2f}'.format(value) for value in values),
                  file=f)

with open('delta.csv', 'w') as f:
    print('# symbol, emt-exp, emt-wien2k, exp-wien2k', file=f)
    for symbol, dct in data.items():
        # Get v0, B, Bp:
        emt, exp, wien2k = [(dct[code + '_volume'],
                             dct[code + '_B'],
                             dct[code + '_Bp'])
                            for code in ['emt', 'exp', 'wien2k']]
        print('{},'.format(symbol),
              '{:.1f}, {:.1f}, {:.1f}'.format(delta(*emt, *exp) * 1000,
                                              delta(*emt, *wien2k) * 1000,
                                              delta(*exp, *wien2k) * 1000),
              file=f)

        if symbol == 'Pt':
            va = min(emt[0], exp[0], wien2k[0])
            vb = max(emt[0], exp[0], wien2k[0])
            v = np.linspace(0.94 * va, 1.06 * vb)
            for (v0, B, Bp), code in [(emt, 'EMT'),
                                      (exp, 'experiment'),
                                      (wien2k, 'WIEN2k')]:
                plt.plot(v, birchmurnaghan(v, 0.0, B, Bp, v0), label=code)
            e0 = dct['emt_energy']
            V = []
            E = []
            for atoms in read('Pt.traj@:'):
                V.append(atoms.get_volume() / len(atoms))
                E.append(atoms.get_potential_energy() / len(atoms) - e0)
            plt.plot(V, E, 'o')
            plt.legend()
            plt.xlabel('volume [Ang^3]')
            plt.ylabel('energy [eV/atom]')
            plt.savefig('Pt.png')



"""Check that QE calculation can run."""

from ase.build import bulk
from ase.calculators.espresso import Espresso

# Default pseudos can go in ~/espresso/pseudo
# Get these from SSSP http://materialscloud.org/sssp/
# Also: Debian/Ubuntu packages provide the below pseudopotential.
# In /usr/share/espresso
PSEUDO = {'Cu': 'Cu.pbe-kjpaw.UPF'}

# Don't forget to
# export ASE_ESPRESSO_COMMAND="mpirun -n 4 $HOME/Compile/q-e/bin/pw.x -in PREFIX.pwi > PREFIX.pwo"
# export ESPRESSO_PSEUDO="/path/to/pseudos"

def main():
    gold = bulk('Cu')
    input_data = {'system':{'occupations': 'smearing',
                            'smearing': 'fermi-dirac',
                            'degauss': 0.02}}
    calc = Espresso(pseudopotentials=PSEUDO,
                    ecutwfc=50.0,
                    input_data=input_data)
    gold.set_calculator(calc)
    gold.get_potential_energy()

    assert calc.get_fermi_level() is not None
    assert calc.get_ibz_k_points() is not None
    assert calc.get_eigenvalues(spin=0, kpt=0) is not None
    assert calc.get_number_of_spins() is not None
    assert calc.get_k_point_weights() is not None

main()



# flake8: noqa
import time

import numpy as np

from ase.transport.tools import dagger
from ase.transport.selfenergy import LeadSelfEnergy
from ase.transport.greenfunction import GreenFunction
from ase.parallel import world


class STM:
    def __init__(self, h1, s1, h2, s2 ,h10, s10, h20, s20, eta1, eta2, w=0.5, pdos=[], logfile = None):
        """XXX
        
        1. Tip
        2. Surface
        
        h1: ndarray
            Hamiltonian and overlap matrix for the isolated tip
            calculation.  Note, h1 should contain (at least) one
            principal layer.

        h2: ndarray
            Same as h1 but for the surface.

        h10: ndarray
            periodic part of the tip. must include two and only
            two principal layers.

        h20: ndarray
            same as h10, but for the surface

        The s* are the corresponding overlap matrices.  eta1, and eta
        2 are (finite) infinitesimals.  """
        
        self.pl1 = len(h10) // 2 #principal layer size for the tip
        self.pl2 = len(h20) // 2 #principal layer size for the surface
        self.h1 = h1 
        self.s1 = s1
        self.h2 = h2
        self.s2 = s2
        self.h10 = h10 
        self.s10 = s10 
        self.h20 = h20 
        self.s20 = s20
        self.eta1 = eta1
        self.eta2 = eta2
        self.w = w #asymmetry of the applied bias (0.5=>symmetric)
        self.pdos = []
        self.log = logfile

    def initialize(self, energies, bias=0):
        """
            energies: list of energies 
            for which the transmission function should be evaluated.
            bias.
            Will precalculate the surface greenfunctions of the tip and
            surface.
        """
        self.bias = bias
        self.energies = energies
        nenergies = len(energies)
        pl1, pl2 = self.pl1, self.pl2
        nbf1, nbf2 = len(self.h1), len(self.h2)
      
        #periodic part of the tip
        hs1_dii = self.h10[:pl1, :pl1], self.s10[:pl1, :pl1]
        hs1_dij = self.h10[:pl1, pl1:2*pl1], self.s10[:pl1, pl1:2*pl1]
        #coupling between per. and non. per part of the tip
        h1_im = np.zeros((pl1, nbf1), complex) 
        s1_im = np.zeros((pl1, nbf1), complex)
        h1_im[:pl1, :pl1], s1_im[:pl1, :pl1] = hs1_dij
        hs1_dim = [h1_im, s1_im]

        #periodic part the surface 
        hs2_dii = self.h20[:pl2, :pl2], self.s20[:pl2, :pl2]
        hs2_dij = self.h20[pl2:2*pl2, :pl2], self.s20[pl2:2*pl2, :pl2]
        #coupling between per. and non. per part of the surface
        h2_im = np.zeros((pl2, nbf2), complex)
        s2_im = np.zeros((pl2, nbf2), complex) 
        h2_im[-pl2:, -pl2:], s2_im[-pl2:, -pl2:] = hs2_dij
        hs2_dim = [h2_im, s2_im]

        #tip and surface greenfunction 
        self.selfenergy1 = LeadSelfEnergy(hs1_dii, hs1_dij, hs1_dim, self.eta1)
        self.selfenergy2 = LeadSelfEnergy(hs2_dii, hs2_dij, hs2_dim, self.eta2)
        self.greenfunction1 = GreenFunction(self.h1-self.bias*self.w*self.s1, self.s1, 
                                            [self.selfenergy1], self.eta1)
        self.greenfunction2 = GreenFunction(self.h2-self.bias*(self.w-1)*self.s2, self.s2, 
                                            [self.selfenergy2], self.eta2)
        
        #Shift the bands due to the bias.
        bias_shift1 = -bias * self.w
        bias_shift2 = -bias * (self.w - 1)
        self.selfenergy1.set_bias(bias_shift1)
        self.selfenergy2.set_bias(bias_shift2)
        
        #tip and surface greenfunction matrices.
        nbf1_small = nbf1 #XXX Change this for efficiency in the future
        nbf2_small = nbf2 #XXX -||-
        coupling_list1 = list(range(nbf1_small))# XXX -||-
        coupling_list2 = list(range(nbf2_small))# XXX -||-
        self.gft1_emm = np.zeros((nenergies, nbf1_small, nbf1_small), complex) 
        self.gft2_emm = np.zeros((nenergies, nbf2_small, nbf2_small), complex)
 
        for e, energy in enumerate(self.energies):
            if self.log != None: # and world.rank == 0:
                    T = time.localtime()
                    self.log.write(' %d:%02d:%02d, ' % (T[3], T[4], T[5]) +
                                   '%d, %d, %02f\n' % (world.rank, e, energy))
            gft1_mm = self.greenfunction1.retarded(energy)[coupling_list1]
            gft1_mm = np.take(gft1_mm, coupling_list1, axis=1)

            gft2_mm = self.greenfunction2.retarded(energy)[coupling_list2]
            gft2_mm = np.take(gft2_mm, coupling_list2, axis=1)
 
            self.gft1_emm[e] = gft1_mm
            self.gft2_emm[e] = gft2_mm

            if self.log != None and world.rank == 0:
                self.log.flush()

    def get_transmission(self, v_12, v_11_2=None, v_22_1=None):
        """XXX

        v_12:
            coupling between tip and surface 
        v_11_2:
            correction to "on-site" tip elements due to the 
            surface (eq.16). Is only included to first order.
        v_22_1:
            corretion to "on-site" surface elements due to he
            tip (eq.17). Is only included to first order.
        """

        dim0 = v_12.shape[0]
        dim1 = v_12.shape[1]

        nenergies = len(self.energies)
        T_e = np.empty(nenergies,float)
        v_21 = dagger(v_12)
        for e, energy in enumerate(self.energies):
            gft1 = self.gft1_emm[e]
            if v_11_2!=None:
                gf1 = np.dot(v_11_2, np.dot(gft1, v_11_2)) 
                gf1 += gft1 #eq. 16
            else:
                gf1 = gft1
            
            gft2 = self.gft2_emm[e]
            if v_22_1!=None:
                gf2 = np.dot(v_22_1,np.dot(gft2, v_22_1))
                gf2 += gft2 #eq. 17
            else:
                gf2 = gft2
            
            a1 = (gf1 - dagger(gf1))
            a2 = (gf2 - dagger(gf2))
            self.v_12 = v_12
            self.a2 = a2
            self.v_21 = v_21
            self.a1 = a1
            v12_a2 = np.dot(v_12, a2[:dim1])
            v21_a1 = np.dot(v_21, a1[-dim0:])
            self.v12_a2 = v12_a2
            self.v21_a1 = v21_a1
            T = -np.trace(np.dot(v12_a2[:,:dim1], v21_a1[:,-dim0:])) #eq. 11
            assert abs(T.imag).max() < 1e-14
            T_e[e] = T.real
            self.T_e = T_e
        return T_e


    def get_current(self, bias, v_12, v_11_2=None, v_22_1=None):
        """Very simple function to calculate the current.
        
        Asummes zero temperature.

        bias: type? XXX
            bias voltage (V)
            
        v_12: XXX
            coupling between tip and surface.
            
        v_11_2:
            correction to onsite elements of the tip
            due to the potential of the surface.
        v_22_1:
            correction to onsite elements of the surface
            due to the potential of the tip.
        """
        energies = self.energies
        T_e = self.get_transmission(v_12, v_11_2, v_22_1)
        bias_window = sorted(-np.array([bias * self.w, bias * (self.w - 1)]))
        self.bias_window = bias_window
        #print 'bias window', np.around(bias_window,3)
        #print 'Shift of tip lead do to the bias:', self.selfenergy1.bias
        #print 'Shift of surface lead do to the bias:', self.selfenergy2.bias
        i1 = sum(energies < bias_window[0]) 
        i2 = sum(energies < bias_window[1])
        step = 1 
        if i2 < i1:
            step = -1
        
        return np.sign(bias)*np.trapz(x=energies[i1:i2:step], y=T_e[i1:i2:step])









import numpy as np

from ase.optimize.optimize import Optimizer


class FIRE(Optimizer):
    def __init__(self, atoms, restart=None, logfile='-', trajectory=None,
                 dt=0.1, maxmove=0.2, dtmax=1.0, Nmin=5, finc=1.1, fdec=0.5,
                 astart=0.1, fa=0.99, a=0.1, master=None, downhill_check=False,
                 position_reset_callback=None, force_consistent=None):
        """Parameters:

        atoms: Atoms object
            The Atoms object to relax.

        restart: string
            Pickle file used to store hessian matrix. If set, file with
            such a name will be searched and hessian matrix stored will
            be used, if the file exists.

        trajectory: string
            Pickle file used to store trajectory of atomic movement.

        logfile: file object or str
            If *logfile* is a string, a file with that name will be opened.
            Use '-' for stdout.

        master: boolean
            Defaults to None, which causes only rank 0 to save files.  If
            set to true,  this rank will save files.

        downhill_check: boolean
            Downhill check directly compares potential energies of subsequent
            steps of the FIRE algorithm rather than relying on the current
            product v*f that is positive if the FIRE dynamics moves downhill.
            This can detect numerical issues where at large time steps the step
            is uphill in energy even though locally v*f is positive, i.e. the
            algorithm jumps over a valley because of a too large time step.

        position_reset_callback: function(atoms, r, e, e_last)
            Function that takes current *atoms* object, an array of position
            *r* that the optimizer will revert to, current energy *e* and
            energy of last step *e_last*. This is only called if e > e_last.

        force_consistent: boolean or None
            Use force-consistent energy calls (as opposed to the energy
            extrapolated to 0 K).  By default (force_consistent=None) uses
            force-consistent energies if available in the calculator, but
            falls back to force_consistent=False if not.  Only meaningful
            when downhill_check is True.
        """
        Optimizer.__init__(self, atoms, restart, logfile, trajectory,
                           master, force_consistent=force_consistent)

        self.dt = dt
        self.Nsteps = 0
        self.maxmove = maxmove
        self.dtmax = dtmax
        self.Nmin = Nmin
        self.finc = finc
        self.fdec = fdec
        self.astart = astart
        self.fa = fa
        self.a = a
        self.downhill_check = downhill_check
        self.position_reset_callback = position_reset_callback

    def initialize(self):
        self.v = None

    def read(self):
        self.v, self.dt = self.load()

    def step(self, f=None):
        atoms = self.atoms

        if f is None:
            f = atoms.get_forces()

        if self.v is None:
            self.v = np.zeros((len(atoms), 3))
            if self.downhill_check:
                self.e_last = atoms.get_potential_energy(
                    force_consistent=self.force_consistent)
                self.r_last = atoms.get_positions().copy()
                self.v_last = self.v.copy()
        else:
            is_uphill = False
            if self.downhill_check:
                e = atoms.get_potential_energy(
                    force_consistent=self.force_consistent)
                # Check if the energy actually decreased
                if e > self.e_last:
                    # If not, reset to old positions...
                    if self.position_reset_callback is not None:
                        self.position_reset_callback(atoms, self.r_last, e,
                                                     self.e_last)
                    atoms.set_positions(self.r_last)
                    is_uphill = True
                self.e_last = atoms.get_potential_energy(
                    force_consistent=self.force_consistent)
                self.r_last = atoms.get_positions().copy()
                self.v_last = self.v.copy()

            vf = np.vdot(f, self.v)
            if vf > 0.0 and not is_uphill:
                self.v = (1.0 - self.a) * self.v + self.a * f / np.sqrt(
                    np.vdot(f, f)) * np.sqrt(np.vdot(self.v, self.v))
                if self.Nsteps > self.Nmin:
                    self.dt = min(self.dt * self.finc, self.dtmax)
                    self.a *= self.fa
                self.Nsteps += 1
            else:
                self.v[:] *= 0.0
                self.a = self.astart
                self.dt *= self.fdec
                self.Nsteps = 0

        self.v += self.dt * f
        dr = self.dt * self.v
        normdr = np.sqrt(np.vdot(dr, dr))
        if normdr > self.maxmove:
            dr = self.maxmove * dr / normdr
        r = atoms.get_positions()
        atoms.set_positions(r + dr)
        self.dump((self.v, self.dt))



from ase.test import cli, require
from ase.db import connect
from ase.io.jsonio import read_json
from ase.io import read
from numpy.testing import assert_allclose

require('nwchem')


cli("""\
ase build O O.traj &&
ase run nwchem O.traj -o nwchem_cmdline.json &&
ase build O2 O2.traj &&
ase run nwchem O2.traj -o nwchem_cmdline.json""")
c = connect('nwchem_cmdline.json')
dct = read_json('nwchem_cmdline.json')
for name in ['O2', 'O']:
    d = c.get([('formula', '=', name)])
    id = d.id
    e1 = d.energy
    e2 = c.get_atoms(id).get_potential_energy()
    e3 = read('{name}.nwo'.format(name=name)).get_potential_energy()
    e4 = dct[id]['energy']
    assert e1 == e2 == e3 == e4
    print(e1)
ae = 2 * c.get('formula=O').energy - c.get('formula=O2').energy
assert_allclose(ae, 6.599194233179787, atol=1e-4, rtol=1e-4)



import re
import time
import numpy as np

from ase.atoms import Atoms
from ase.utils import reader, writer
from ase.cell import Cell

__all__ = ['read_rmc6f', 'write_rmc6f']

ncols2style = {9: 'no_labels',
               10: 'labels',
               11: 'magnetic'}


def _read_construct_regex(lines):
    """
    Utility for constructing  regular expressions used by reader.
    """
    lines = [l.strip() for l in lines]
    lines_re = '|'.join(lines)
    lines_re = lines_re.replace(' ', r'\s+')
    lines_re = lines_re.replace('(', r'\(')
    lines_re = lines_re.replace(')', r'\)')
    return '({})'.format(lines_re)


def _read_line_of_atoms_section(fields):
    """
    Process `fields` line of Atoms section in rmc6f file and output extracted
    info as atom id (key) and list of properties for Atoms object (values).

    Parameters
    ----------
    fields: list[str]
        List of columns from line in rmc6f file.


    Returns
    ------
    atom_id: int
        Atom ID
    properties: list[str|float]
        List of Atoms properties based on rmc6f style.
        Basically, have 1) element and fractional coordinates for 'labels'
        or 'no_labels' style and 2) same for 'magnetic' style except adds
        the spin.
        Examples for 1) 'labels' or 'no_labels' styles or 2) 'magnetic' style:
            1) [element, xf, yf, zf]
            2) [element, xf, yf, zf, spin]
    """
    # Atom id
    atom_id = int(fields[0])

    # Get style used for rmc6f from file based on number of columns
    ncols = len(fields)
    style = ncols2style[ncols]

    # Create the position dictionary
    properties = list()
    element = str(fields[1])
    if style == 'no_labels':
        # id element xf yf zf ref_num ucell_x ucell_y ucell_z
        xf = float(fields[2])
        yf = float(fields[3])
        zf = float(fields[4])
        properties = [element, xf, yf, zf]

    elif style == 'labels':
        # id element label xf yf zf ref_num ucell_x ucell_y ucell_z
        xf = float(fields[3])
        yf = float(fields[4])
        zf = float(fields[5])
        properties = [element, xf, yf, zf]

    elif style == 'magnetic':
        # id element xf yf zf ref_num ucell_x ucell_y ucell_z M: spin
        xf = float(fields[2])
        yf = float(fields[3])
        zf = float(fields[4])
        spin = float(fields[10].strip("M:"))
        properties = [element, xf, yf, zf, spin]
    else:
        raise Exception("Unsupported style for parsing rmc6f file format.")

    return atom_id, properties


def _read_process_rmc6f_lines_to_pos_and_cell(lines):
    """
    Processes the lines of rmc6f file to atom position dictionary and cell

    Parameters
    ----------
    lines: list[str]
        List of lines from rmc6f file.

    Returns
    ------
    pos : dict{int:list[str|float]}
        Dict for each atom id and Atoms properties based on rmc6f style.
        Basically, have 1) element and fractional coordinates for 'labels'
        or 'no_labels' style and 2) same for 'magnetic' style except adds
        the spin.
        Examples for 1) 'labels' or 'no_labels' styles or 2) 'magnetic' style:
            1) pos[aid] = [element, xf, yf, zf]
            2) pos[aid] = [element, xf, yf, zf, spin]
    cell: Cell object
        The ASE Cell object created from cell parameters read from the 'Cell'
        section of rmc6f file.
    """

    # Inititalize output pos dictionary
    pos = {}

    # Defined the header an section lines we process
    header_lines = [
        "Number of atoms:",
        "Supercell dimensions:",
        "Cell (Ang/deg):",
        "Lattice vectors (Ang):"]
    sections = ["Atoms"]

    # Construct header and sections regex
    header_lines_re = _read_construct_regex(header_lines)
    sections_re = _read_construct_regex(sections)

    section = None
    header = True

    # Remove any lines that are blank
    lines = [line for line in lines if line != '']

    # Process each line of rmc6f file
    pos = {}
    for line in lines:

        # check if in a section
        m = re.match(sections_re, line)
        if m is not None:
            section = m.group(0).strip()
            header = False
            continue

        # header section
        if header:
            field = None
            val = None

            # Regex that matches whitespace-separated floats
            float_list_re = r'\s+(\d[\d|\s\.]+[\d|\.])'
            m = re.search(header_lines_re + float_list_re, line)
            if m is not None:
                field = m.group(1)
                val = m.group(2)

            if field is not None and val is not None:

                if field == "Number of atoms:":
                    pass
                    """
                    NOTE: Can just capture via number of atoms ingested.
                          Maybe use in future for a check.
                    code: natoms = int(val)
                    """

                if field.startswith('Supercell'):
                    pass
                    """
                    NOTE: wrapping back down to unit cell is not
                          necessarily needed for ASE object.

                    code: supercell = [int(x) for x in val.split()]
                    """

                if field.startswith('Cell'):
                    cellpar = [float(x) for x in val.split()]
                    cell = Cell.fromcellpar(cellpar)

                if field.startswith('Lattice'):
                    pass
                    """
                    NOTE: Have questions about RMC fractionalization matrix for
                          conversion in data2config vs. the ASE matrix.
                          Currently, just support the Cell section.
                    """

        # main body section
        if section is not None:
            if section == 'Atoms':
                atom_id, atom_props = _read_line_of_atoms_section(line.split())
                pos[atom_id] = atom_props

    return pos, cell


def _write_output_column_format(columns, arrays):
    """
    Helper function to build output for data columns in rmc6f format

    Parameters
    ----------
    columns: list[str]
        List of keys in arrays. Will be columns in the output file.
    arrays: dict{str:np.array}
        Dict with arrays for each column of rmc6f file that are
        property of Atoms object.

    Returns
    ------
    property_ncols : list[int]
        Number of columns for each property.
    dtype_obj: np.dtype
        Data type object for the columns.
    formats_as_str: str
        The format for printing the columns.

    """
    fmt_map = {'d': ('R', '%14.6f '),
               'f': ('R', '%14.6f '),
               'i': ('I', '%8d '),
               'O': ('S', '%s'),
               'S': ('S', '%s'),
               'U': ('S', '%s'),
               'b': ('L', ' %.1s ')}

    property_types = []
    property_ncols = []
    dtypes = []
    formats = []

    # Take each column and setup formatting vectors
    for column in columns:
        array = arrays[column]
        dtype = array.dtype

        property_type, fmt = fmt_map[dtype.kind]
        property_types.append(property_type)

        # Flags for 1d vectors
        is_1d = len(array.shape) == 1
        is_1d_as_2d = len(array.shape) == 2 and array.shape[1] == 1

        # Construct the list of key pairs of column with dtype
        if (is_1d or is_1d_as_2d):
            ncol = 1
            dtypes.append((column, dtype))
        else:
            ncol = array.shape[1]
            for c in range(ncol):
                dtypes.append((column + str(c), dtype))

        # Add format and number of columns for this property to output array
        formats.extend([fmt] * ncol)
        property_ncols.append(ncol)

    # Prepare outputs to correct data structure
    dtype_obj = np.dtype(dtypes)
    formats_as_str = ''.join(formats) + '\n'

    return property_ncols, dtype_obj, formats_as_str


def _write_output(filename, header_lines, data, fmt, order=None):
    """
    Helper function to write information to the filename

    Parameters
    ----------
    filename : file|str
        A file like object or filename
    header_lines : list[str]
        Header section of output rmc6f file
    data: np.array[len(atoms)]
        Array for the Atoms section to write to file. Has
        the columns that need to be written on each row
    fmt: str
        The format string to use for writing each column in
        the rows of data.
    order : list[str]
        If not None, gives a list of atom types for the order
        to write out each.
    """
    f = filename

    # Write header section
    for line in header_lines:
        f.write("%s \n" % line)

    # If specifying the order, fix the atom id and write to file
    natoms = data.shape[0]
    if order is not None:
        new_id = 0
        for atype in order:
            for i in range(natoms):
                if atype == data[i][1]:
                    new_id += 1
                    data[i][0] = new_id
                    f.write(fmt % tuple(data[i]))
    # ...just write rows to file
    else:
        for i in range(natoms):
            f.write(fmt % tuple(data[i]))


@reader
def read_rmc6f(filename, atom_type_map=None):
    """
    Parse a RMCProfile rmc6f file into ASE Atoms object

    Parameters
    ----------
    filename : file|str
        A file like object or filename.
    atom_type_map: dict{str:str}
        Map of atom types for conversions. Mainly used if there is
        an atom type in the file that is not supported by ASE but
        want to map to a supported atom type instead.

        Example to map deuterium to hydrogen:
        atom_type_map = { 'D': 'H' }

    Returns
    ------
    structure : Atoms
        The Atoms object read in from the rmc6f file.
    """

    f = filename
    lines = f.readlines()

    # Process the rmc6f file to extract positions and cell
    pos, cell = _read_process_rmc6f_lines_to_pos_and_cell(lines)

    # create an atom type map if one does not exist from unique atomic symbols
    if atom_type_map is None:
        symbols = [atom[0] for atom in pos.values()]
        atom_type_map = {atype: atype for atype in symbols}

    # Use map of tmp symbol to actual symbol
    for atom in pos.values():
        atom[0] = atom_type_map[atom[0]]

    # create Atoms from read-in data
    symbols = []
    scaled_positions = []
    spin = None
    magmoms = []
    for atom in pos.values():
        if len(atom) == 4:
            element, x, y, z = atom
        else:
            element, x, y, z, spin = atom

        element = atom_type_map[element]
        symbols.append(element)
        scaled_positions.append([x, y, z])
        if spin is not None:
            magmoms.append(spin)

    atoms = Atoms(scaled_positions=scaled_positions,
                  symbols=symbols,
                  cell=cell,
                  magmoms=magmoms,
                  pbc=[True, True, True])

    return atoms


@writer
def write_rmc6f(filename, atoms, order=None, atom_type_map=None):
    """
    Write output in rmc6f format - RMCProfile v6 fractional coordinates

    Parameters
    ----------
    filename : file|str
        A file like object or filename.
    atoms: Atoms object
        The Atoms object to be written.

    order : list[str]
        If not None, gives a list of atom types for the order
        to write out each.
    atom_type_map: dict{str:str}
        Map of atom types for conversions. Mainly used if there is
        an atom type in the Atoms object that is a placeholder
        for a different atom type. This is used when the atom type
        is not supported by ASE but is in RMCProfile.

        Example to map hydrogen to deuterium:
        atom_type_map = { 'H': 'D' }
    """

    # get atom types and how many of each (and set to order if passed)
    atom_types = set(atoms.symbols)
    if order is not None:
        if set(atom_types) != set(order):
            raise Exception("The order is not a set of the atom types.")
        atom_types = order

    atom_count_dict = atoms.symbols.formula.count()
    natom_types = [str(atom_count_dict[atom_type]) for atom_type in atom_types]

    # create an atom type map if one does not exist from unique atomic symbols
    if atom_type_map is None:
        symbols = set(np.array(atoms.symbols))
        atom_type_map = {atype: atype for atype in symbols}

    # HEADER SECTION

    # get type and number of each atom type
    atom_types_list = [atom_type_map[atype] for atype in atom_types]
    atom_types_present = ' '.join(atom_types_list)
    natom_types_present = ' '.join(natom_types)

    header_lines = [
        "(Version 6f format configuration file)",
        "(Generated by ASE - Atomic Simulation Environment https://wiki.fysik.dtu.dk/ase/ )",  # noqa: E501
        "Metadata date: " + time.strftime('%d-%m-%Y'),
        "Number of types of atoms:   {} ".format(len(atom_types)),
        "Atom types present:          {}".format(atom_types_present),
        "Number of each atom type:   {}".format(natom_types_present),
        "Number of moves generated:           0",
        "Number of moves tried:               0",
        "Number of moves accepted:            0",
        "Number of prior configuration saves: 0",
        "Number of atoms:                     {}".format(len(atoms)),
        "Supercell dimensions:                1 1 1"]

    # magnetic moments
    if atoms.has('magmoms'):
        spin_str = "Number of spins:                     {}"
        spin_line = spin_str.format(len(atoms.get_initial_magnetic_moments()))
        header_lines.extend([spin_line])

    density_str = "Number density (Ang^-3):              {}"
    density_line = density_str.format(len(atoms) / atoms.get_volume())
    cell_angles = [str(x) for x in atoms.get_cell_lengths_and_angles()]
    cell_line = "Cell (Ang/deg): " + ' '.join(cell_angles)
    header_lines.extend([density_line, cell_line])

    # get lattice vectors from cell lengths and angles
    # NOTE: RMCProfile uses a different convention for the fractionalization
    # matrix

    cell_parameters = atoms.get_cell_lengths_and_angles()
    cell = Cell.fromcellpar(cell_parameters).T
    x_line = ' '.join(['{:12.6f}'.format(i) for i in cell[0]])
    y_line = ' '.join(['{:12.6f}'.format(i) for i in cell[1]])
    z_line = ' '.join(['{:12.6f}'.format(i) for i in cell[2]])
    lat_lines = ["Lattice vectors (Ang):", x_line, y_line, z_line]
    header_lines.extend(lat_lines)
    header_lines.extend(['Atoms:'])

    # ATOMS SECTION

    # create columns of data for atoms (fr_cols)
    fr_cols = ['id', 'symbols', 'scaled_positions', 'ref_num', 'ref_cell']
    if atoms.has('magmoms'):
        fr_cols.extend('magmom')

    # Collect data to be written out
    natoms = len(atoms)

    arrays = {}
    arrays['id'] = np.array(range(1, natoms + 1, 1), int)
    arrays['symbols'] = np.array(atoms.symbols)
    arrays['ref_num'] = np.zeros(natoms, int)
    arrays['ref_cell'] = np.zeros((natoms, 3), int)
    arrays['scaled_positions'] = np.array(atoms.get_scaled_positions())

    # get formatting for writing output based on atom arrays
    ncols, dtype_obj, fmt = _write_output_column_format(fr_cols, arrays)

    # Pack fr_cols into record array
    data = np.zeros(natoms, dtype_obj)
    for column, ncol in zip(fr_cols, ncols):
        value = arrays[column]
        if ncol == 1:
            data[column] = np.squeeze(value)
        else:
            for c in range(ncol):
                data[column + str(c)] = value[:, c]

    # Use map of tmp symbol to actual symbol
    for i in range(natoms):
        data[i][1] = atom_type_map[data[i][1]]

    # Write the output
    _write_output(filename, header_lines, data, fmt, order=order)



import numpy as np
from ase.calculators.calculator import Calculator
from ase import units

k_c = units.Hartree * units.Bohr


class AtomicCounterIon(Calculator):
    implemented_properties = ['energy', 'forces']

    def __init__(self, charge, epsilon, sigma, sites_per_mol=1,
                 rc=7.0, width=1.0):
        """ Counter Ion Calculator.

        A very simple, nonbonded (Coulumb and LJ)
        interaction calculator meant for single atom ions
        to charge neutralize systems (and nothing else)...
        """
        self.rc = rc
        self.width = width
        self.sites_per_mol = sites_per_mol
        self.epsilon = epsilon
        self.sigma = sigma
        self.charge = charge
        Calculator.__init__(self)

    def add_virtual_sites(self, positions):
        return positions

    def get_virtual_charges(self, atoms):
        charges = np.tile(self.charge, len(atoms) // self.sites_per_mol)
        return charges

    def redistribute_forces(self, forces):
        return forces

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        R = atoms.get_positions()
        charges = self.get_virtual_charges(atoms)
        pbc = atoms.pbc

        energy = 0.0
        forces = np.zeros_like(atoms.get_positions())

        for m in range(len(atoms)):
            D = R[m + 1:] - R[m]
            shift = np.zeros_like(D)
            for i, periodic in enumerate(pbc):
                if periodic:
                    L = atoms.cell.diagonal()[i]
                    shift[:, i] = (D[:, i] + L / 2) % L - L / 2 - D[:, i]
            D += shift
            d2 = (D**2).sum(1)
            d = d2**0.5

            x1 = d > self.rc - self.width
            x2 = d < self.rc
            x12 = np.logical_and(x1, x2)
            y = (d[x12] - self.rc + self.width) / self.width
            t = np.zeros(len(d))  # cutoff function
            t[x2] = 1.0
            t[x12] -= y**2 * (3.0 - 2.0 * y)
            dtdd = np.zeros(len(d))
            dtdd[x12] -= 6.0 / self.width * y * (1.0 - y)

            c6 = (self.sigma**2 / d2)**3
            c12 = c6**2
            e_lj = 4 * self.epsilon * (c12 - c6)
            e_c = k_c * charges[m + 1:] * charges[m] / d

            energy += np.dot(t, e_lj)
            energy += np.dot(t, e_c)

            F = (24 * self.epsilon * (2 * c12 - c6) / d2 * t -
                 e_lj * dtdd / d)[:, None] * D

            forces[m] -= F.sum(0)
            forces[m + 1:] += F

            F = (e_c / d2 * t)[:, None] * D \
                - (e_c * dtdd / d)[:, None] * D

            forces[m] -= F.sum(0)
            forces[m + 1:] += F

        self.results['energy'] = energy
        self.results['forces'] = forces




# creates:  diffusion-path.png
from ase.io import read, write
if 1:
    exec(compile(open('diffusion4.py').read(), 'diffusion4.py', 'exec'))
images = [read('mep%d.traj' % i) for i in range(5)]
a = images[0] + images[1] + images[2] + images[3] + images[4]
del a.constraints
a *= (2, 1, 1)
a.set_cell(images[0].get_cell())
write('diffusion-path.pov', a, rotation='-90x',
      transparent=False, run_povray=True)



from ase.build import molecule
from ase.calculators.emt import EMT
from ase.optimize import QuasiNewton
from ase.vibrations import Vibrations
from ase.thermochemistry import IdealGasThermo

atoms = molecule('N2')
atoms.set_calculator(EMT())
dyn = QuasiNewton(atoms)
dyn.run(fmax=0.01)
potentialenergy = atoms.get_potential_energy()

vib = Vibrations(atoms)
vib.run()
vib_energies = vib.get_energies()

thermo = IdealGasThermo(vib_energies=vib_energies,
                        potentialenergy=potentialenergy,
                        atoms=atoms,
                        geometry='linear',
                        symmetrynumber=2, spin=0)
G = thermo.get_gibbs_energy(temperature=298.15, pressure=101325.)



import numpy as np

from ase.io import read
from ase.io.jsonio import read_json
from ase.geometry import crystal_structure_from_cell
from ase.dft.kpoints import (get_special_points, special_paths, BandPath,
                             parse_path_string, labels_from_kpts,
                             get_monkhorst_pack_size_and_offset)

def atoms2bandpath(atoms, path='default',
                   k_points=False,
                   ibz_k_points=False,
                   dimension=3,
                   verbose=False):
    cell = atoms.get_cell()
    icell = atoms.get_reciprocal_cell()

    try:
        cs = crystal_structure_from_cell(cell)
    except ValueError:
        cs = None

    if verbose:
        if cs:
            print('Crystal:', cs)
            print('Special points:', special_paths[cs])
        print('Lattice vectors:')
        for i, v in enumerate(cell):
            print('{}: ({:16.9f},{:16.9f},{:16.9f})'.format(i + 1, *v))
        print('Reciprocal vectors:')
        for i, v in enumerate(icell):
            print('{}: ({:16.9f},{:16.9f},{:16.9f})'.format(i + 1, *v))

    # band path
    special_points = None
    if path:
        if path == 'default':
            path = special_paths[cs]
        paths = []
        special_points = get_special_points(cell)
        for names in parse_path_string(path):
            points = []
            for name in names:
                points.append(np.dot(icell.T, special_points[name]))
            paths.append((names, points))
    else:
        paths = None

    # k points
    points = None
    if atoms.calc is not None and hasattr(atoms.calc, 'get_bz_k_points'):
        bzk = atoms.calc.get_bz_k_points()
        if path is None:
            try:
                size, offset = get_monkhorst_pack_size_and_offset(bzk)
            except ValueError:
                # This was not a MP-grid.  Must be a path in the BZ:
                path = ''.join(labels_from_kpts(bzk, cell)[2])

        if k_points:
            points = bzk
        elif ibz_k_points:
            points = atoms.calc.get_ibz_k_points()

    return BandPath(cell, kpts=points, special_points=special_points)


def plot_reciprocal_cell(path, output=None, dimension=3,
                         plot_vectors=True):
    import matplotlib.pyplot as plt

    path.plot(dimension=dimension, vectors=plot_vectors)

    if output:
        plt.savefig(output)
    else:
        plt.show()


class CLICommand:
    """Show the reciprocal space.

    Read unit cell from a file and show a plot of the 1. Brillouin zone.  If
    the file contains information about k-points, then those can be plotted
    too.

    Examples:

        $ # Show GXWLG path in FCC-BZ:
        $ ase build -x fcc Al al.traj
        $ ase reciprocal al.traj -p GXWLG

        $ # And now with k-points:
        $ ase run gpaw al.traj -p kpts=6,6,6,mode=pw \
        >   --after "atoms.calc.write('al.gpw')" > al.txt
        $ ase reciprocal al.gpw -i -p GXWLG
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('name', metavar='input-file',
            help='Input file containing unit cell.')
        add('output', nargs='?', help='Write plot to file (.png, .svg, ...).')
        add('-v', '--verbose', action='store_true', help='More output.')
        add('-p', '--path', nargs='?', type=str, const='default',
            help='Add a band path.  Example: "GXL".')
        add('-d', '--dimension', type=int, default=3,
            help='Dimension of the cell.')
        add('--no-vectors', action='store_true',
            help="Don't show reciprocal vectors.")
        kp = parser.add_mutually_exclusive_group(required=False)
        kp.add_argument('-k', '--k-points', action='store_true',
                        help='Add k-points of the calculator.')
        kp.add_argument('-i', '--ibz-k-points', action='store_true',
                        help='Add irreducible k-points of the calculator.')

    @staticmethod
    def run(args, parser):
        from ase.io.formats import UnknownFileTypeError

        try:
            atoms = read(args.name)
        except UnknownFileTypeError:
            # Probably a bandpath/bandstructure:
            obj = read_json(args.name)
            if isinstance(obj, BandPath):
                path = obj
            elif hasattr(obj, 'path'):  # Probably band structure
                path = obj.path
            else:
                parser.error('Strange object: {}'.format(obj))
        else:
            path = atoms2bandpath(atoms, path=args.path,
                                  verbose=args.verbose,
                                  k_points=args.k_points,
                                  ibz_k_points=args.ibz_k_points)

        plot_reciprocal_cell(path,
                             output=args.output,
                             dimension=args.dimension,
                             plot_vectors=not args.no_vectors)



from ase import Atoms
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
from ase.optimize import QuasiNewton
from ase.build import fcc111, add_adsorbate

h = 1.85
d = 1.10

slab = fcc111('Cu', size=(4, 4, 2), vacuum=10.0)

slab.set_calculator(EMT())
e_slab = slab.get_potential_energy()

molecule = Atoms('2N', positions=[(0., 0., 0.), (0., 0., d)])
molecule.set_calculator(EMT())
e_N2 = molecule.get_potential_energy()

add_adsorbate(slab, molecule, h, 'ontop')
constraint = FixAtoms(mask=[a.symbol != 'N' for a in slab])
slab.set_constraint(constraint)
dyn = QuasiNewton(slab, trajectory='N2Cu.traj')
dyn.run(fmax=0.05)

print('Adsorption energy:', e_slab + e_N2 - slab.get_potential_energy())



import os
import unittest

from ase import Atoms
from ase.calculators.acemolecule import ACE

label = "test"
mol = Atoms('H2',[(0, 0, 0),(0, 0, 0.7)])
basic = [dict(Cell= '5.0')]
if "ASE_ACE_COMMAND" not in os.environ:
    raise unittest.SkipTest('$ASE_ACE_COMMAND not defined') 
ace = ACE(label=label, BasicInformation=basic)
mol.set_calculator(ace)
mol.get_forces()
#mol.get_potential_energy()



""" Test AtomicCounterIon is force/energy consistent over 
    PBCs and with cutoff """

import numpy as np
from ase import Atoms
from ase import units
from ase.calculators.counterions import AtomicCounterIon as ACI

sigma = 1.868 * (1.0/2.0)**(1.0/6.0)
epsilon = 0.00277 * units.kcal / units.mol

atoms = Atoms('3Na', positions=np.array([[0, 0, -2], [0,0,0], [0,0,2]]))
atoms.cell = [10, 10, 10]
atoms.pbc = True

atoms.calc = ACI(1, epsilon, sigma, rc=4.5)
points = np.arange(-15., 15., 0.2)

for p in points:
    f = atoms.get_forces()
    fn = atoms.calc.calculate_numerical_forces(atoms, 1e-5)
    df = (f-fn)
    assert abs(df).max() < 1e-8



from functools import reduce
from itertools import combinations, chain
from math import factorial
from operator import mul

import numpy as np

from ase.units import kg, C, _hbar, kB
from ase.vibrations import Vibrations


class Factorial:
    def __init__(self):
        self._fac = [1]
        self._inv = [1.]

    def __call__(self, n):
        try:
            return self._fac[n]
        except IndexError:
            for i in range(len(self._fac), n + 1):
                self._fac.append(i * self._fac[i - 1])
                try:
                    self._inv.append(float(1. / self._fac[-1]))
                except OverflowError:
                    self._inv.append(0.)
            return self._fac[n]

    def inv(self, n):
        self(n)
        return self._inv[n]


class FranckCondonOverlap:
    """Evaluate squared overlaps depending on the Huang-Rhys parameter."""
    def __init__(self):
        self.factorial = Factorial()

    def directT0(self, n, S):
        """|<0|n>|^2

        Direct squared Franck-Condon overlap corresponding to T=0.
        """
        return np.exp(-S) * S**n * self.factorial.inv(n)

    def direct(self, n, m, S_in):
        """|<n|m>|^2

        Direct squared Franck-Condon overlap.
        """
        if n > m:
            # use symmetry
            return self.direct(m, n, S_in)

        S = np.array([S_in])
        mask = np.where(S == 0)
        S[mask] = 1  # hide zeros
        s = 0
        for k in range(n + 1):
            s += (-1)**(n - k) * S**float(-k) / (
                self.factorial(k) *
                self.factorial(n - k) * self.factorial(m - k))
        res = np.exp(-S) * S**(n + m) * s**2 * (
            self.factorial(n) * self.factorial(m))
        # use othogonality
        res[mask] = int(n == m)
        return res[0]

    def direct0mm1(self, m, S):
        """<0|m><m|1>"""
        sum = S**m
        if m:
            sum -= m * S**(m - 1)
        return np.exp(-S) * np.sqrt(S) * sum * self.factorial.inv(m)

    def direct0mm2(self, m, S):
        """<0|m><m|2>"""
        sum = S**(m + 1)
        if m >= 1:
            sum -= 2 * m * S**m
        if m >= 2:
            sum += m * (m - 1) * S**(m - 1)
        return np.exp(-S) / np.sqrt(2) * sum * self.factorial.inv(m)


class FranckCondonRecursive:
    """Recursive implementation of Franck-Condon overlaps

    Notes
    -----
    The ovelaps are signed according to the sign of the displacements.

    Reference
    ---------
    Julien Guthmuller
    The Journal of Chemical Physics 144, 064106 (2016); doi: 10.1063/1.4941449
    """
    def __init__(self):
        self.factorial = Factorial()

    def ov0m(self, m, delta):
        if m == 0:
            return np.exp(-0.25 * delta**2)
        else:
            assert(m > 0)
            return - delta / np.sqrt(2 * m) * self.ov0m(m - 1, delta)
            
    def ov1m(self, m, delta):
        sum = delta * self.ov0m(m, delta) / np.sqrt(2.)
        if m == 0:
            return sum
        else:
            assert(m > 0)
            return sum + np.sqrt(m) * self.ov0m(m - 1, delta)
            
    def ov2m(self, m, delta):
        sum = delta * self.ov1m(m, delta) / 2
        if m == 0:
            return sum
        else:
            assert(m > 0)
            return sum + np.sqrt(m / 2.) * self.ov1m(m - 1, delta)
            
    def ov3m(self, m, delta):
        sum = delta * self.ov2m(m, delta) / np.sqrt(6.)
        if m == 0:
            return sum
        else:
            assert(m > 0)
            return sum + np.sqrt(m / 3.) * self.ov2m(m - 1, delta)
            
    def ov0mm1(self, m, delta):
        if m == 0:
            return delta / np.sqrt(2) * self.ov0m(m, delta)**2
        else:
            return delta / np.sqrt(2) * (
                self.ov0m(m, delta)**2 - self.ov0m(m - 1, delta)**2)
            
    def direct0mm1(self, m, delta):
        """direct and fast <0|m><m|1>"""
        S = delta**2 / 2.
        sum = S**m
        if m:
            sum -= m * S**(m - 1)
        return np.where(S == 0, 0,
                        (np.exp(-S) * delta / np.sqrt(2) * sum *
                         self.factorial.inv(m)))

    def ov0mm2(self, m, delta):
        if m == 0:
            return delta**2 / np.sqrt(8) * self.ov0m(m, delta)**2
        elif m == 1:
            return delta**2 / np.sqrt(8) * (
                self.ov0m(m, delta)**2 - 2 * self.ov0m(m - 1, delta)**2)
        else:
            return delta**2 / np.sqrt(8) * (
                self.ov0m(m, delta)**2 - 2 * self.ov0m(m - 1, delta)**2 +
                self.ov0m(m - 2, delta)**2)

    def direct0mm2(self, m, delta):
        """direct and fast <0|m><m|2>"""
        S = delta**2 / 2.
        sum = S**(m + 1)
        if m >= 1:
            sum -= 2 * m * S**m
        if m >= 2:
            sum += m * (m - 1) * S**(m - 1)
        return np.exp(-S) / np.sqrt(2) * sum * self.factorial.inv(m)

    def ov1mm2(self, m, delta):
        p1 = delta**3 / 4.
        sum = p1 * self.ov0m(m, delta)**2
        if m == 0:
            return sum
        p2 = delta - 3. * delta**3 / 4
        sum += p2 * self.ov0m(m - 1, delta)**2
        if m == 1:
            return sum
        sum -= p2 * self.ov0m(m - 2, delta)**2
        if m == 2:
            return sum
        return sum - p1 * self.ov0m(m - 3, delta)**2

    def direct1mm2(self, m, delta):
        S = delta**2 / 2.
        sum = S**2
        if m > 0:
            sum -= 2 * m * S
        if m > 1:
            sum += m * (m - 1)
        with np.errstate(divide='ignore', invalid='ignore'):
            return np.where(S == 0, 0,
                (np.exp(-S) * S**(m - 1) / delta * (S - m) * sum *
                 self.factorial.inv(m)))

    def direct0mm3(self, m, delta):
        S = delta**2 / 2.
        with np.errstate(divide='ignore', invalid='ignore'):
            return np.where(S == 0, 0,
                (np.exp(-S) * S**(m - 1) / delta * np.sqrt(12.) *
                 (S**3 / 6. - m * S**2 / 2 +
                  m * (m - 1) * S / 2. - m * (m - 1) * (m - 2) / 6) *
                 self.factorial.inv(m)))


class FranckCondon:
    def __init__(self, atoms, vibname, minfreq=-np.inf, maxfreq=np.inf):
        """Input is a atoms object and the corresponding vibrations.
        With minfreq and maxfreq frequencies can
        be excluded from the calculation"""

        self.atoms = atoms
        # V = a * v is the combined atom and xyz-index
        self.mm05_V = np.repeat(1. / np.sqrt(atoms.get_masses()), 3)
        self.minfreq = minfreq
        self.maxfreq = maxfreq
        self.shape = (len(self.atoms), 3)

        vib = Vibrations(atoms, name=vibname)
        self.energies = np.real(vib.get_energies(method='frederiksen'))  # [eV]
        self.frequencies = np.real(
            vib.get_frequencies(method='frederiksen'))  # [cm^-1]
        self.modes = vib.modes
        self.H = vib.H

    def get_Huang_Rhys_factors(self, forces):
        """Evaluate Huang-Rhys factors and corresponding frequencies
        from forces on atoms in the exited electronic state.
        The double harmonic approximation is used. HR factors are
        the first approximation of FC factors,
        no combinations or higher quanta (>1) exitations are considered"""

        assert(forces.shape == self.shape)

        # Hesse matrix
        H_VV = self.H
        # sqrt of inverse mass matrix
        mm05_V = self.mm05_V
        # mass weighted Hesse matrix
        Hm_VV = mm05_V[:, None] * H_VV * mm05_V
        # mass weighted displacements
        Fm_V = forces.flat * mm05_V
        X_V = np.linalg.solve(Hm_VV, Fm_V)
        # projection onto the modes
        modes_VV = self.modes
        d_V = np.dot(modes_VV, X_V)
        # Huang-Rhys factors S
        s = 1.e-20 / kg / C / _hbar**2  # SI units
        S_V = s * d_V**2 * self.energies / 2

        # reshape for minfreq
        indices = np.where(self.frequencies <= self.minfreq)
        np.append(indices, np.where(self.frequencies >= self.maxfreq))
        S_V = np.delete(S_V, indices)
        frequencies = np.delete(self.frequencies, indices)

        return S_V, frequencies

    def get_Franck_Condon_factors(self, order, temp, forces):
        """Return FC factors and corresponding frequencies up to given order.

        order= number of quanta taken into account
        T= temperature in K. Vibronic levels are occupied by a
        Boltzman distribution.
        forces= forces on atoms in the exited electronic state"""

        S, f = self.get_Huang_Rhys_factors(forces)
        n = order + 1
        T = temp
        freq = np.array(f)

        # frequencies
        freq_n = [[] * i for i in range(n - 1)]
        freq_neg = [[] * i for i in range(n - 1)]

        for i in range(1, n):
            freq_n[i - 1] = freq * i
            freq_neg[i - 1] = freq * (-i)

        # combinations
        freq_nn = [x for x in combinations(chain(*freq_n), 2)]
        for i in range(len(freq_nn)):
            freq_nn[i] = freq_nn[i][0] + freq_nn[i][1]

        indices2 = []
        for i, y in enumerate(freq):
            ind = [j for j, x in enumerate(freq_nn) if x % y == 0]
            indices2.append(ind)
        indices2 = [x for x in chain(*indices2)]
        freq_nn = np.delete(freq_nn, indices2)

        frequencies = [[] * x for x in range(3)]
        frequencies[0].append(freq_neg[0])
        frequencies[0].append([0])
        frequencies[0].append(freq_n[0])
        frequencies[0] = [x for x in chain(*frequencies[0])]

        for i in range(1, n - 1):
            frequencies[1].append(freq_neg[i])
            frequencies[1].append(freq_n[i])
        frequencies[1] = [x for x in chain(*frequencies[1])]

        frequencies[2] = freq_nn

        # Franck-Condon factors
        E = freq / 8065.5
        f_n = [[] * i for i in range(n)]

        for j in range(0, n):
            f_n[j] = np.exp(-E * j / (kB * T))

        # partition function
        Z = np.empty(len(S))
        Z = np.sum(f_n, 0)

        # occupation probability
        w_n = [[] * k for k in range(n)]
        for l in range(n):
            w_n[l] = f_n[l] / Z

        # overlap wavefunctions
        O_n = [[] * m for m in range(n)]
        O_neg = [[] * m for m in range(n)]
        for o in range(n):
            O_n[o] = [[] * p for p in range(n)]
            O_neg[o] = [[] * p for p in range(n - 1)]
            for q in range(o, n + o):
                a = np.minimum(o, q)
                summe = []
                for k in range(a + 1):
                    s = ((-1)**(q - k) * np.sqrt(S)**(o + q - 2 * k) *
                         factorial(o) * factorial(q) /
                         (factorial(k) * factorial(o - k) * factorial(q - k)))
                    summe.append(s)
                summe = np.sum(summe, 0)
                O_n[o][q - o] = (np.exp(-S / 2) /
                                 (factorial(o) * factorial(q))**(0.5) *
                                 summe)**2 * w_n[o]
            for q in range(n - 1):
                O_neg[o][q] = [0 * b for b in range(len(S))]
            for q in range(o - 1, -1, -1):
                a = np.minimum(o, q)
                summe = []
                for k in range(a + 1):
                    s = ((-1)**(q - k) * np.sqrt(S)**(o + q - 2 * k) *
                         factorial(o) * factorial(q) /
                         (factorial(k) * factorial(o - k) * factorial(q - k)))
                    summe.append(s)
                summe = np.sum(summe, 0)
                O_neg[o][q] = (np.exp(-S / 2) /
                               (factorial(o) * factorial(q))**(0.5) *
                               summe)**2 * w_n[o]
        O_neg = np.delete(O_neg, 0, 0)

        # Franck-Condon factors
        FC_n = [[] * i for i in range(n)]
        FC_n = np.sum(O_n, 0)
        zero = reduce(mul, FC_n[0])
        FC_neg = [[] * i for i in range(n - 2)]
        FC_neg = np.sum(O_neg, 0)
        FC_n = np.delete(FC_n, 0, 0)

        # combination FC factors
        FC_nn = [x for x in combinations(chain(*FC_n), 2)]
        for i in range(len(FC_nn)):
            FC_nn[i] = FC_nn[i][0] * FC_nn[i][1]

        FC_nn = np.delete(FC_nn, indices2)

        FC = [[] * x for x in range(3)]
        FC[0].append(FC_neg[0])
        FC[0].append([zero])
        FC[0].append(FC_n[0])
        FC[0] = [x for x in chain(*FC[0])]

        for i in range(1, n - 1):
            FC[1].append(FC_neg[i])
            FC[1].append(FC_n[i])
        FC[1] = [x for x in chain(*FC[1])]

        FC[2] = FC_nn

        """Returned are two 3-dimensional lists. First inner list contains
frequencies and FC-factors of vibrations exited with |1| quanta and
the 0-0 transition.
        Second list contains frequencies and FC-factors from higher
quanta exitations. Third list are combinations of two normal modes
(including combinations of higher quanta exitations). """
        return FC, frequencies



import numpy as np
from numpy.random import RandomState
from ase.phonons import Phonons
from ase.data import atomic_numbers
from ase.optimize import FIRE
#from asap3 import EMT
from ase.calculators.emt import EMT
from ase.build import bulk
from ase.md.velocitydistribution import PhononHarmonics
from ase import units

# Tests the phonon-based perturbation and velocity distribution
# for thermal equilibration in MD.

rng = RandomState(17)

atoms = bulk('Pd')
atoms *= (3, 3, 3)
avail = [atomic_numbers[sym]
         for sym in ['Ni', 'Cu', 'Pd', 'Ag', 'Pt', 'Au']]
atoms.numbers[:] = rng.choice(avail, size=len(atoms))
atoms.calc = EMT()

opt = FIRE(atoms, trajectory='relax.traj')
opt.run(fmax=0.001)
positions0 = atoms.positions.copy()

phonons = Phonons(atoms, EMT(), supercell=(1, 1, 1), delta=0.05)

try:
    phonons.run()
    phonons.read()  # Why all this boilerplate?
finally:
    phonons.clean()
matrices = phonons.get_force_constant()

K = matrices[0]
T = 300 * units.kB

atoms.calc = EMT()
Epotref = atoms.get_potential_energy()

temps = []
Epots = []
Ekins = []
Etots = []


for i in range(24):
    PhononHarmonics(atoms, K, T, quantum=True, rng=np.random.RandomState(888 + i))

    Epot = atoms.get_potential_energy() - Epotref
    Ekin = atoms.get_kinetic_energy()
    Ekins.append(Ekin)
    Epots.append(Epot)
    Etots.append(Ekin + Epot)
    temps.append(atoms.get_temperature())

    atoms.positions[:] = positions0

    # The commented code would produce displacements/velocities
    # resolved over phonon modes if we borrow some expressions
    # from the function.  Each mode should contribute on average
    # equally to both Epot and Ekin/temperature
    #
    # atoms1.calc = EMT()
    # atoms1 = atoms.copy()
    # v_ac = np.zeros_like(positions0)
    # D_acs, V_acs = ...
    # for s in range(V_acs.shape[2]):
    #     atoms1.positions += D_acs[:, :, s]
    #     v_ac += V_acs[:, :, s]
    #     atoms1.set_velocities(v_ac)
    #     X1.append(atoms1.get_potential_energy() - Epotref)
    #     X2.append(atoms1.get_kinetic_energy())

    print('energies', Epot, Ekin, Epot + Ekin)



Epotmean = np.mean(Epots)
Ekinmean = np.mean(Ekins)
Tmean = np.mean(temps)
Terr = abs(Tmean - T / units.kB)
relative_imbalance = abs(Epotmean - Ekinmean) / (Epotmean + Ekinmean)


print('epotmean', Epotmean)
print('ekinmean', Ekinmean)
print('rel imbalance', relative_imbalance)
print('Tmean', Tmean, 'Tref', T / units.kB, 'err', Terr)

assert Terr < 0.1*T / units.kB, Terr  # error in Kelvin for instantaneous velocity
# Epot == Ekin give or take 2 %:
assert relative_imbalance < 0.1, relative_imbalance


if 0:
    import matplotlib.pyplot as plt
    I = np.arange(len(Epots))
    plt.plot(I, Epots, 'o', label='pot')
    plt.plot(I, Ekins, 'o', label='kin')
    plt.plot(I, Etots, 'o', label='tot')
    plt.show()




# flake8: noqa
'''
mathematical utilities
'''
import numpy as np
import bisect

def vinterp3d(x,y,z,u,xi,yi,zi):

    np.array([xi,yi,zi])

    #1D arrays of cooridinates
    xv = x[:,0,0]
    yv = y[0,:,0]
    zv = z[0,0,:]

    # we subtract 1 because bisect tells us where to insert the
    # element to maintain an ordered list, so we want the index to the
    # left of that point
    i = bisect.bisect_right(xv,xi) - 1
    j = bisect.bisect_right(yv,yi) - 1
    k = bisect.bisect_right(zv,zi) - 1

    #occasionally we get the edge of the cell and we then go one index
    #back again
    if i == len(xv)-1: i-=1
    if j == len(yv)-1: j-=1
    if k == len(zv)-1: k-=1


    #points at edge of cell. We only need P1, P2, P3, and P5
    P1 = np.array([x[i,j,k],y[i,j,k],z[i,j,k]])
    P2 = np.array([x[i+1,j,k],y[i+1,j,k],z[i+1,j,k]])
    P3 = np.array([x[i,j+1,k],y[i,j+1,k],z[i,j+1,k]])
    P5 = np.array([x[i,j,k+1],y[i,j,k+1],z[i,j,k+1]])

    #values of u at edge of cell
    u1 = u[i,j,k]
    u2 = u[i+1,j,k]
    u3 = u[i,j+1,k]
    u4 = u[i+1,j+1,k]
    u5 = u[i,j,k+1]
    u6 = u[i+1,j,k+1]
    u7 = u[i,j+1,k+1]
    u8 = u[i+1,j+1,k+1]

    #cell basis vectors, not the unit cell, but the voxel cell containing the point
    cbasis = np.array([P2-P1,
                       P3-P1,
                       P5-P1])

    #now get interpolated point in terms of the cell basis
    s = np.dot(np.linalg.inv(cbasis.T),np.array([xi,yi,zi])-P1)

    #now s = (sa, sb, sc) which are fractional coordinates in the vector space
    #next we do the interpolations
    ui1 = u1 + s[0]*(u2-u1)
    ui2 = u3 + s[0]*(u4-u3)

    ui3 = u5 + s[0]*(u6-u5)
    ui4 = u7 + s[0]*(u8-u7)

    ui5 = ui1 + s[1]*(ui2-ui1)
    ui6 = ui3 + s[1]*(ui4-ui3)

    ui7 = ui5 + s[2]*(ui6-ui5)

    return ui7




"""Checkpointing and restart functionality for scripts using ASE Atoms objects.

Initialize checkpoint object:

CP = Checkpoint('checkpoints.db')

Checkpointed code block in try ... except notation:

try:
    a, C, C_err = CP.load()
except NoCheckpoint:
    C, C_err = fit_elastic_constants(a)
    CP.save(a, C, C_err)

Checkpoint code block, shorthand notation:

C, C_err = CP(fit_elastic_constants)(a)

Example for checkpointing within an iterative loop, e.g. for searching crack
tip position:

try:
    a, converged, tip_x, tip_y = CP.load()
except NoCheckpoint:
    converged = False
    tip_x = tip_x0
    tip_y = tip_y0
while not converged:
    ... do something to find better crack tip position ...
    converged = ...
    CP.flush(a, converged, tip_x, tip_y)

The simplest way to use checkpointing is through the CheckpointCalculator. It
wraps any calculator object and does a checkpoint whenever a calculation
is performed:

    calc = ...
    cp_calc = CheckpointCalculator(calc)
    atoms.set_calculator(cp_calc)
    e = atoms.get_potential_energy() # 1st time, does calc, writes to checkfile
                                     # subsequent runs, reads from checkpoint
"""

import numpy as np

import ase
from ase.db import connect
from ase.calculators.calculator import Calculator


class NoCheckpoint(Exception):
    pass

    
class DevNull:
    def write(str, *args):
        pass

        
class Checkpoint(object):
    _value_prefix = '_values_'

    def __init__(self, db='checkpoints.db', logfile=None):
        self.db = db
        if logfile is None:
            logfile = DevNull()
        self.logfile = logfile

        self.checkpoint_id = [0]
        self.in_checkpointed_region = False

    def __call__(self, func, *args, **kwargs):
        checkpoint_func_name = str(func)
        
        def decorated_func(*args, **kwargs):
            # Get the first ase.Atoms object.
            atoms = None
            for a in args:
                if atoms is None and isinstance(a, ase.Atoms):
                    atoms = a

            try:
                retvals = self.load(atoms=atoms)
            except NoCheckpoint:
                retvals = func(*args, **kwargs)
                if isinstance(retvals, tuple):
                    self.save(*retvals, atoms=atoms,
                              checkpoint_func_name=checkpoint_func_name)
                else:
                    self.save(retvals, atoms=atoms,
                              checkpoint_func_name=checkpoint_func_name)
            return retvals
        return decorated_func

    def _increase_checkpoint_id(self):
        if self.in_checkpointed_region:
            self.checkpoint_id += [1]
        else:
            self.checkpoint_id[-1] += 1
        self.logfile.write('Entered checkpoint region '
                           '{0}.\n'.format(self.checkpoint_id))

        self.in_checkpointed_region = True

    def _decrease_checkpoint_id(self):
        self.logfile.write('Leaving checkpoint region '
                           '{0}.\n'.format(self.checkpoint_id))
        if not self.in_checkpointed_region:
            self.checkpoint_id = self.checkpoint_id[:-1]
            assert len(self.checkpoint_id) >= 1
        self.in_checkpointed_region = False
        assert self.checkpoint_id[-1] >= 1

    def _mangled_checkpoint_id(self):
        """
        Returns a mangled checkpoint id string:
            check_c_1:c_2:c_3:...
        E.g. if checkpoint is nested and id is [3,2,6] it returns:
            'check3:2:6'
        """
        return 'check'+':'.join(str(id) for id in self.checkpoint_id)

    def load(self, atoms=None):
        """
        Retrieve checkpoint data from file. If atoms object is specified, then
        the calculator connected to that object is copied to all returning
        atoms object.

        Returns tuple of values as passed to flush or save during checkpoint
        write.
        """
        self._increase_checkpoint_id()

        retvals = []
        with connect(self.db) as db:
            try:
                dbentry = db.get(checkpoint_id=self._mangled_checkpoint_id())
            except KeyError:
                raise NoCheckpoint

            data = dbentry.data
            atomsi = data['checkpoint_atoms_args_index']
            i = 0
            while (i == atomsi or
                   '{0}{1}'.format(self._value_prefix, i) in data):
                if i == atomsi:
                    newatoms = dbentry.toatoms()
                    if atoms is not None:
                        # Assign calculator
                        newatoms.set_calculator(atoms.get_calculator())
                    retvals += [newatoms]
                else:
                    retvals += [data['{0}{1}'.format(self._value_prefix, i)]]
                i += 1

        self.logfile.write('Successfully restored checkpoint '
                           '{0}.\n'.format(self.checkpoint_id))
        self._decrease_checkpoint_id()
        if len(retvals) == 1:
            return retvals[0]
        else:
            return tuple(retvals)

    def _flush(self, *args, **kwargs):
        data = dict(('{0}{1}'.format(self._value_prefix, i), v)
                    for i, v in enumerate(args))

        try:
            atomsi = [isinstance(v, ase.Atoms) for v in args].index(True)
            atoms = args[atomsi]
            del data['{0}{1}'.format(self._value_prefix, atomsi)]
        except ValueError:
            atomsi = -1
            try:
                atoms = kwargs['atoms']
            except KeyError:
                raise RuntimeError('No atoms object provided in arguments.')

        try:
            del kwargs['atoms']
        except KeyError:
            pass

        data['checkpoint_atoms_args_index'] = atomsi
        data.update(kwargs)

        with connect(self.db) as db:
            try:
                dbentry = db.get(checkpoint_id=self._mangled_checkpoint_id())
                del db[dbentry.id]
            except KeyError:
                pass
            db.write(atoms, checkpoint_id=self._mangled_checkpoint_id(),
                     data=data)

        self.logfile.write('Successfully stored checkpoint '
                           '{0}.\n'.format(self.checkpoint_id))

    def flush(self, *args, **kwargs):
        """
        Store data to a checkpoint without increasing the checkpoint id. This
        is useful to continously update the checkpoint state in an iterative
        loop.
        """
        # If we are flushing from a successfully restored checkpoint, then
        # in_checkpointed_region will be set to False. We need to reset to True
        # because a call to flush indicates that this checkpoint is still
        # active.
        self.in_checkpointed_region = False
        self._flush(*args, **kwargs)

    def save(self, *args, **kwargs):
        """
        Store data to a checkpoint and increase the checkpoint id. This closes
        the checkpoint.
        """
        self._decrease_checkpoint_id()
        self._flush(*args, **kwargs)


def atoms_almost_equal(a, b, tol=1e-9):
    return (np.abs(a.positions - b.positions).max() < tol and
            (a.numbers == b.numbers).all() and
            np.abs(a.cell - b.cell).max() < tol and
            (a.pbc == b.pbc).all())


class CheckpointCalculator(Calculator):
    """
    This wraps any calculator object to checkpoint whenever a calculation
    is performed.

    This is particularily useful for expensive calculators, e.g. DFT and
    allows usage of complex workflows.

    Example usage:

        calc = ...
        cp_calc = CheckpointCalculator(calc)
        atoms.set_calculator(cp_calc)
        e = atoms.get_potential_energy()
        # 1st time, does calc, writes to checkfile
        # subsequent runs, reads from checkpoint file
    """
    implemented_properties = ase.calculators.calculator.all_properties
    default_parameters = {}
    name = 'CheckpointCalculator'

    property_to_method_name = {
        'energy': 'get_potential_energy',
        'energies': 'get_potential_energies',
        'forces': 'get_forces',
        'stress': 'get_stress',
        'stresses': 'get_stresses'}

    def __init__(self, calculator, db='checkpoints.db', logfile=None):
        Calculator.__init__(self)
        self.calculator = calculator
        if logfile is None:
            logfile = DevNull()
        self.checkpoint = Checkpoint(db, logfile)
        self.logfile = logfile

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        try:
            results = self.checkpoint.load(atoms)
            prev_atoms, results = results[0], results[1:]
            try:
                assert atoms_almost_equal(atoms, prev_atoms)
            except AssertionError:
                raise AssertionError('mismatch between current atoms and '
                                     'those read from checkpoint file')
            self.logfile.write('retrieved results for {0} from checkpoint\n'
                               .format(properties))
            # save results in calculator for next time
            if isinstance(self.calculator, Calculator):
                if not hasattr(self.calculator, 'results'):
                    self.calculator.results = {}
                self.calculator.results.update(dict(zip(properties, results)))
        except NoCheckpoint:
            if isinstance(self.calculator, Calculator):
                self.logfile.write('doing calculation of {0} with new-style '
                                   'calculator interface\n'.format(properties))
                self.calculator.calculate(atoms, properties, system_changes)
                results = [self.calculator.results[prop]
                           for prop in properties]
            else:
                self.logfile.write('doing calculation of {0} with old-style '
                                   'calculator interface\n'.format(properties))
                results = []
                for prop in properties:
                    method_name = self.property_to_method_name[prop]
                    method = getattr(self.calculator, method_name)
                    results.append(method(atoms))
            _calculator = atoms.get_calculator()
            try:
                atoms.set_calculator(self.calculator)
                self.checkpoint.save(atoms, *results)
            finally:
                atoms.set_calculator(_calculator)

        self.results = dict(zip(properties, results))



"""

Check the unit cell is handled correctly

"""

from ase.test.vasp import installed2 as installed
from ase.calculators.vasp import Vasp2 as Vasp
from ase.build import molecule
from ase.test import must_raise
assert installed()


# Molecules come with no unit cell

atoms = molecule('CH4')
calc = Vasp()

with must_raise(ValueError):
    atoms.set_calculator(calc)
    atoms.get_total_energy()



import itertools
from ase.cell import Cell


def all_pbcs():
    values = [False, True]
    yield from itertools.product(values, values, values)


def test(cell):
    for pbc in all_pbcs():
        ucell = cell.uncomplete(pbc)
        assert ucell.rank == sum(pbc & cell.any(1))

    assert all(cell.uncomplete(True).any(1) == cell.any(1)), (cell.uncomplete(True), cell)
    assert all(cell.uncomplete(1).any(1) == cell.any(1))
    assert cell.uncomplete(False).rank == 0
    assert cell.uncomplete(0).rank == 0

test(Cell.new([3, 4, 5]))
test(Cell.new([2, 0, 3]))



from ase.atoms import Atoms
from ase.constraints import FixScaled

a1 = Atoms(symbols = 'X2',
           positions = [[0.,0.,0.], [2.,0.,0.], ],
           cell = [[4.,0.,0.], [0.,4.,0.], [0.,0.,4.], ],
           )

fs1 = FixScaled(a1.get_cell(), -1, mask=(True, False, False))
fs2 = FixScaled(a1.get_cell(), 1, mask=(False, True, False))

a1.set_constraint([fs1,fs2])

# reassigning using atoms.__getitem__
a2 = a1[0:2]

assert len(a1._constraints) == len(a2._constraints)




from ase.build import bulk
from ase.calculators.emt import EMT
from ase.utils.forcecurve import force_curve
from ase.md import VelocityVerlet
from ase.units import fs
from ase.io import read

atoms = bulk('Au', cubic=True) * (2, 1, 1)
atoms.calc = EMT()
atoms.rattle(stdev=0.05)

md = VelocityVerlet(atoms, timestep=12.0 * fs, trajectory='tmp.traj')
md.run(steps=10)
images = read('tmp.traj', ':')
force_curve(images)

# import pylab as plt
# plt.show()



import numpy as np

from ase import Atoms
from ase.units import Bohr, Ry


def read_scf(filename):
    try:
        f = open(filename + '.scf', 'r')
        pip = f.readlines()
        ene = []
        for line in pip:
            if line[0:4] == ':ENE':
                ene.append(float(line[43:59]) * Ry)
        f.close()
        return ene
    except:
        return None

        
def read_struct(filename, ase=True):
    f = open(filename, 'r')
    pip = f.readlines()
    lattice = pip[1][0:3]
    nat = int(pip[1][27:30])
    cell = np.zeros(6)
    for i in range(6):
        cell[i] = float(pip[3][0 + i * 10:10 + i * 10])
    cell[0:3] = cell[0:3] * Bohr
    if lattice == 'P  ':
        lattice = 'P'
    elif lattice == 'H  ':
        lattice = 'P'
        cell[3:6] = [90.0, 90.0, 120.0]
    elif lattice == 'R  ':
        lattice = 'R'
    elif lattice == 'F  ':
        lattice = 'F'
    elif lattice == 'B  ':
        lattice = 'I'
    elif lattice == 'CXY':
        lattice = 'C'
    elif lattice == 'CXZ':
        lattice = 'B'
    elif lattice == 'CYZ':
        lattice = 'A'
    else:
        raise RuntimeError('TEST needed')
    pos = np.array([])
    atomtype = []
    rmt = []
    neq = np.zeros(nat)
    iline = 4
    indif = 0
    for iat in range(nat):
        indifini = indif
        if len(pos) == 0:
            pos = np.array([[float(pip[iline][12:22]),
                             float(pip[iline][25:35]),
                             float(pip[iline][38:48])]])
        else:
            pos = np.append(pos, np.array([[float(pip[iline][12:22]),
                                            float(pip[iline][25:35]),
                                            float(pip[iline][38:48])]]),
                            axis=0)
        indif += 1
        iline += 1
        neq[iat] = int(pip[iline][15:17])
        iline += 1
        for ieq in range(1, int(neq[iat])):
            pos = np.append(pos, np.array([[float(pip[iline][12:22]),
                                            float(pip[iline][25:35]),
                                            float(pip[iline][38:48])]]),
                            axis=0)
            indif += 1
            iline += 1
        for i in range(indif - indifini):
            atomtype.append(pip[iline][0:2].replace(' ', ''))
            rmt.append(float(pip[iline][43:48]))
        iline += 4
    if ase:
        cell2 = coorsys(cell)
        atoms = Atoms(atomtype, pos, pbc=True)
        atoms.set_cell(cell2, scale_atoms=True)
        cell2 = np.dot(c2p(lattice), cell2)
        if lattice == 'R':
            atoms.set_cell(cell2, scale_atoms=True)
        else:
            atoms.set_cell(cell2)
        return atoms
    else:
        return cell, lattice, pos, atomtype, rmt

        
def write_struct(filename, atoms2=None, rmt=None, lattice='P', zza=None):
    atoms = atoms2.copy()
    atoms.wrap()
    f = open(filename, 'w')
    f.write('ASE generated\n')
    nat = len(atoms)
    if rmt is None:
        rmt = [2.0] * nat
        f.write(lattice +
                '   LATTICE,NONEQUIV.ATOMS:%3i\nMODE OF CALC=RELA\n' % nat)
    cell = atoms.get_cell()
    metT = np.dot(cell, np.transpose(cell))
    cell2 = cellconst(metT)
    cell2[0:3] = cell2[0:3] / Bohr
    f.write(('%10.6f' * 6) % tuple(cell2) + '\n')
    if zza is None:
        zza = atoms.get_atomic_numbers()
    for ii in range(nat):
        f.write('ATOM %3i: ' % (ii + 1))
        pos = atoms.get_scaled_positions()[ii]
        f.write('X=%10.8f Y=%10.8f Z=%10.8f\n' % tuple(pos))
        f.write('          MULT= 1          ISPLIT= 1\n')
        zz = zza[ii]
        if zz > 71:
            ro = 0.000005
        elif zz > 36:
            ro = 0.00001
        elif zz > 18:
            ro = 0.00005
        else:
            ro = 0.0001
        f.write('%-10s NPT=%5i  R0=%9.8f RMT=%10.4f   Z:%10.5f\n' %
                (atoms.get_chemical_symbols()[ii], 781, ro, rmt[ii], zz))
        f.write('LOCAL ROT MATRIX:    %9.7f %9.7f %9.7f\n' % (1.0, 0.0, 0.0))
        f.write('                     %9.7f %9.7f %9.7f\n' % (0.0, 1.0, 0.0))
        f.write('                     %9.7f %9.7f %9.7f\n' % (0.0, 0.0, 1.0))
    f.write('   0\n')

    
def cellconst(metT):
    """ metT=np.dot(cell,cell.T) """
    aa = np.sqrt(metT[0, 0])
    bb = np.sqrt(metT[1, 1])
    cc = np.sqrt(metT[2, 2])
    gamma = np.arccos(metT[0, 1] / (aa * bb)) / np.pi * 180.0
    beta = np.arccos(metT[0, 2] / (aa * cc)) / np.pi * 180.0
    alpha = np.arccos(metT[1, 2] / (bb * cc)) / np.pi * 180.0
    return np.array([aa, bb, cc, alpha, beta, gamma])

    
def coorsys(latconst):
    a = latconst[0]
    b = latconst[1]
    c = latconst[2]
    cal = np.cos(latconst[3] * np.pi / 180.0)
    cbe = np.cos(latconst[4] * np.pi / 180.0)
    cga = np.cos(latconst[5] * np.pi / 180.0)
    sga = np.sin(latconst[5] * np.pi / 180.0)
    return np.array([[a, b * cga, c * cbe],
                     [0, b * sga, c * (cal - cbe * cga) / sga],
                     [0, 0, c * np.sqrt(1 - cal**2 - cbe**2 - cga**2 +
                                        2 * cal * cbe * cga) / sga]
                     ]).transpose()

    
def c2p(lattice):
    """ apply as eg. cell2 = np.dot(c2p('F'), cell)"""
    if lattice == 'P':
        cell = np.eye(3)
    elif lattice == 'F':
        cell = np.array([[0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]])
    elif lattice == 'I':
        cell = np.array([[-0.5, 0.5, 0.5], [0.5, -0.5, 0.5], [0.5, 0.5, -0.5]])
    elif lattice == 'C':
        cell = np.array([[0.5, 0.5, 0.0], [0.5, -0.5, 0.0], [0.0, 0.0, -1.0]])
    elif lattice == 'B':
        cell = np.array([[0.5, 0.0, 0.5], [0.0, 1.0, 0.0], [0.5, 0.0, -0.5]])
    elif lattice == 'A':
        cell = np.array([[-1.0, 0.0, 0.0], [0.0, -0.5, 0.5], [0.0, 0.5, 0.5]])
    elif lattice == 'R':
        cell = np.array([[2.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0],
                         [-1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0],
                         [-1.0 / 3.0, -2.0 / 3.0, 1.0 / 3.0]])

    else:
        raise ValueError('lattice is ' + lattice + '!')
    return cell



from ase import io
from ase.calculators.vdwcorrection import vdWTkatchenko09prl
from ase.calculators.emt import EMT
from ase.build import bulk

# fake objects for the test
class FakeHirshfeldPartitioning:
    def __init__(self, calculator):
        self.calculator = calculator
    def initialize(self):
        pass
    def get_effective_volume_ratios(self):
        return [1]
    def get_calculator(self):
        return self.calculator
class FakeDFTcalculator(EMT):
    def get_xc_functional(self):
        return 'PBE'

a = 4.05  # Angstrom lattice spacing
al = bulk('Al', 'fcc', a=a)

cc = FakeDFTcalculator()
hp = FakeHirshfeldPartitioning(cc)
c = vdWTkatchenko09prl(hp, [3])
al.set_calculator(c)
al.get_potential_energy()

fname = 'out.traj'
al.write(fname)

# check that the output exist
atoms = io.read(fname)
p = io.read(fname).get_calculator().parameters
p['calculator']
p['xc']
p['uncorrected_energy']



import os
import warnings

import numpy as np
from ase import Atoms
from ase.io import write, read, iread
from ase.io.formats import all_formats, ioformats
from ase.calculators.singlepoint import SinglePointCalculator

try:
    import matplotlib
except ImportError:
    matplotlib = 0

try:
    from lxml import etree
except ImportError:
    etree = 0

try:
    import Scientific
except ImportError:
    Scientific = 0

try:
    import netCDF4
except ImportError:
    netCDF4 = 0


def get_atoms():
    a = 5.0
    d = 1.9
    c = a / 2
    atoms = Atoms('AuH',
                  positions=[(0, c, c), (d, c, c)],
                  cell=(2 * d, a, a),
                  pbc=(1, 0, 0))
    extra = np.array([2.3, 4.2])
    atoms.set_array('extra', extra)
    atoms *= (2, 1, 1)

    # attach some results to the Atoms.
    # These are serialised by the extxyz writer.

    spc = SinglePointCalculator(atoms,
                                energy=-1.0,
                                stress=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                                forces=-1.0 * atoms.positions)
    atoms.set_calculator(spc)
    return atoms


def check(a, ref_atoms, format):
    assert abs(a.positions - ref_atoms.positions).max() < 1e-6, \
        (a.positions - ref_atoms.positions)
    if format in ['traj', 'cube', 'cfg', 'struct', 'gen', 'extxyz',
                  'db', 'json', 'trj']:
        assert abs(a.cell - ref_atoms.cell).max() < 1e-6
    if format in ['cfg', 'extxyz']:
        assert abs(a.get_array('extra') -
                   ref_atoms.get_array('extra')).max() < 1e-6
    if format in ['extxyz', 'traj', 'trj', 'db', 'json']:
        assert (a.pbc == ref_atoms.pbc).all()
        assert a.get_potential_energy() == ref_atoms.get_potential_energy()
        assert (a.get_stress() == ref_atoms.get_stress()).all()
        assert abs(a.get_forces() - ref_atoms.get_forces()).max() < 1e-12


testdir = 'tmp_io_testdir'
if os.path.isdir(testdir):
    import shutil
    shutil.rmtree(testdir)

os.mkdir(testdir)


def test(format):
    if format in ['abinit', 'castep-cell', 'dftb', 'eon', 'gaussian',
                  'lammps-data']:
        # Someone should do something ...
        return

    if format in ['v-sim', 'mustem']:
        # Standalone test used as not compatible with 1D periodicity
        return

    if format in ['mustem']:
        # Standalone test used as specific arguments are required
        return

    if format in ['dmol-arc', 'dmol-car', 'dmol-incoor']:
        # We have a standalone dmol test
        return

    if format in ['gif', 'mp4']:
        # Complex dependencies; see animate.py test
        return

    if format in ['postgresql', 'trj', 'vti', 'vtu', 'mysql']:
        # Let's not worry about these.
        return

    if not matplotlib and format in ['eps', 'png']:
        return

    if not etree and format == 'exciting':
        return

    if not Scientific and format == 'etsf':
        return

    if not netCDF4 and format == 'netcdftrajectory':
        return

    atoms = get_atoms()

    if format == 'dlp4':
        atoms.pbc = (1, 1, 0)

    images = [atoms, atoms]

    io = ioformats[format]
    print('{0:20}{1}{2}{3}{4}'.format(format,
                                      ' R'[bool(io.read)],
                                      ' W'[bool(io.write)],
                                      '+1'[io.single],
                                      'SF'[io.acceptsfd]))
    fname1 = '{}/io-test.1.{}'.format(testdir, format)
    fname2 = '{}/io-test.2.{}'.format(testdir, format)
    if io.write:
        write(fname1, atoms, format=format)
        if not io.single:
            write(fname2, images, format=format)

        if io.read:
            for a in [read(fname1, format=format), read(fname1)]:
                check(a, atoms, format)

            if not io.single:
                if format in ['json', 'db']:
                    aa = read(fname2 + '@id=1') + read(fname2 + '@id=2')
                else:
                    aa = [read(fname2), read(fname2, 0)]
                aa += read(fname2, ':')
                for a in iread(fname2, format=format):
                    aa.append(a)
                assert len(aa) == 6, aa
                for a in aa:
                    check(a, atoms, format)

for format in sorted(all_formats):
    with warnings.catch_warnings():
        if format in ['proteindatabank', 'netcdftrajectory']:
            warnings.simplefilter('ignore', UserWarning)
        test(format)
